# qfit.py
# based on dat_reader.py and data_optimizer.py by JB@KIT, 
# 2015-2018 by JB@KIT and the qkit team
# data reading and fitting

# import and basic usage

'''
usage:

.. code-block:: python

    from qkit import qfit as qfit
    qf = qfit.QFIT()
    qf.load('data/file.h5', entries=['time', 'phase'])
    qf.rotate_IQ_plane()
    qf.fit_exp()

'''
# a customized fit funcitoncan be set via directly writing in self.fit_function
# and providing full initial parameters
# for further information see doc strings

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

#\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=

import numpy as np
plot_enable = False
try:
    import qkit
    if qkit.module_available("matplotlib"):
        import matplotlib.pyplot as plt
        plot_enable = True
except (ImportError, AttributeError):
    try:
        import matplotlib.pyplot as plt
        plot_enable = True
    except ImportError:
        plot_enable = False
from scipy.optimize import curve_fit
import scipy.interpolate
import os, glob
import time
import logging

try:
    from qkit.storage import store
except ImportError: pass

#\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=

def _fill_p0(guesses,p0):
    '''
    Fill estimated guesses with specified initial values (in p0). 
    p0 can be a smaller list than len(guesses) and may contain 'None' entries 
    e.g. guesses=[0,0,0,0,0] and p0 = [None,1,2] then the resulting guesses = [0,1,2,0,0]
    '''
    if guesses is None: return p0
    if p0 is not None:
        try:
            for i,p in enumerate(p0):
                if p is not None:
                    guesses[i] = p
        except Exception:
            logging.error('List of given initial parameters invalid. Aborting.')
            raise ValueError
    return guesses

#\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=

class QFIT(object):
    def __init__(self):

        self.cfg = {'show_plot' : True,
                    'show_output' : True,
                    'save_png' : True,
                    'save_pdf' : False,
                    'show_complex_plot' : True,             #show complex plot during optimize
                    'data_column' : 2,                      #data column entry if entry list is longer than 2
                    'spline_knots' : 1.e-3,                 #spacing between knots used during spline fitting, smaller number = more accurate
                    'figsize' : (15,7),                     #figure size setting
                    'debug' : False,                        #debug mode
                    'analysis_folder' : 'analysis',         #folder in h5 file where data is to be stored
                    'datafolder_structure' : 'day_time'}    #day_time or epoche_key

        try:
            import qkit
            self.cfg['data_dir'] = qkit.cfg.get('datadir').replace('\\','/')
            if qkit.cfg.get('datafolder_structure',1) == 2:   #new data structure with epoche key in qkit settings
                try:
                    self.run_id = qkit.cfg['run_id']
                    self.user = qkit.cfg['user']
                    self.cfg['datafolder_structure'] = 'epoche_key'
                except KeyError:
                    logging.error('No run_id and user entry found in qkit.cfg. Falling back to old style data folder structure.')
            else:
                self.cfg['datafolder_structure'] = 'day_time'
        except (ImportError, AttributeError):
            logging.warning('Running outside of qkit environment.')
            self.cfg['data_dir'] = os.getcwd()

        try:
            store
            self.cfg['store'] = True
        except ImportError:
            logging.warning('Package store not found.')
            self.cfg['store'] = False
        self.cfg['matplotlib'] = plot_enable

    # fit function definitions --------------------------------------------------------------------------------------

    def __f_Lorentzian_sqrt(self, f, f0, k, a, offs):
        return np.sign(a) * np.sqrt(np.abs(a**2*(k/2)**2/((k/2)**2+(f-f0)**2)))+offs
        
    def __f_Lorentzian(self, f, f0, k, a, offs):
        return a*k/(2*np.pi)/((k/2)**2+(f-f0)**2)+offs

    def __f_Skewed_Lorentzian(self, f, f0, k, a, offs, tilt, skew):
        return (a + skew * (f - f0)) * k / (2 * np.pi) / ((k / 2) ** 2 + (f - f0) ** 2) + offs + tilt * (f - f0)

    #def __f_general_reflection(f, f0, kc, ki, a, offs):
    #    return a*(4*kc*(4*(f-f0)**2+kc**2-ki**2))/(16*(f-f0)**4+8*(f-f0)**2*kc**2+8*(f-f0)**2*ki**2+kc**4-2*kc**2*ki**2+ki**4)+offs
        
    def __f_damped_sine(self, t, fs, Td, a, offs, ph):
        if a < 0: return np.NaN #constrict amplitude to positive values
        if fs < 0: return np.NaN
        if Td < 0: return np.NaN
        #if ph < -np.pi or ph > np.pi: return np.NaN #constrict phase
        return a*np.exp(-t/Td)*np.sin(2*np.pi*fs*t+ph)+offs
        
    def __f_sine(self, t, fs, a, offs, ph):
        if a < 0: return np.NaN #constrict amplitude to positive values
        return a*np.sin(2*np.pi*fs*t+ph)+offs
        
    def __f_exp(self, t, Td, a, offs):
        return a*np.exp(-t/Td)+offs
        
    def __f_exp_sine(self, t, fs, Td, a, offs, ph, d):
        return a*np.exp(-t/Td)*0.5*(1+d*np.cos(2*np.pi*fs*t+ph))+offs

    def __get_parameters(self, fit_function):
        '''
        Parameters of known fit functions used for plotting purposes.
        '''
        return {self.__f_Lorentzian_sqrt: ['f0','k','a','offs'],
            self.__f_Lorentzian: ['f0','k','a','offs'],
            self.__f_Skewed_Lorentzian:['f0','k','a','offs','tilt','skew'],
            self.__f_sine: ['f','a','offs','ph'],
            self.__f_exp: ['Td','a','offs'],
            self.__f_damped_sine: ['f','Td','a','offs','ph','d']
        }.get(fit_function)

    #\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=

    def load(self, *args, **kwargs):
        '''
        Load data from either:
         - numpy arrays: must specify keyword arguments 'coordinate' and 'data'
         - numpy arrays: must specify keyword arguments 'coordinate' and 'amplitude' and 'phase'
         - data file (h5 or text based file): specify filename
         - recent data file in data_dir: no arguments provided

        h5 entries need to be given with keyword 'entries'.
        '''

        if 'coordinate' in kwargs.keys() and 'data' in kwargs.keys():
            self.coordinate = kwargs['coordinate']
            self.data = kwargs['data']
            self.amplitude = None
            self.phase = None
            self.urls = None
            self.file_name = 'data_import'
            self.coordinate_label = ''
            self.data_label = ''
            return
        elif 'coordinate' in kwargs.keys() and 'amplitude' in kwargs.keys() and 'phase' in  kwargs.keys():
            #print("loading amplitude and phase signal from np array")
            self.coordinate = kwargs['coordinate']
            self.data = None
            self.amplitude = kwargs['amplitude']
            self.phase = kwargs['phase']
            self.urls = None
            self.file_name = 'data_import'
            self.coordinate_label = ''
            self.data_label = ''
            return
        elif len(args) == 0:   #no file name provided
            if self.cfg['store']: ftype = 'h5'
            else: ftype = 'dat'
            if self.find_latest_file(ftype) == 0:
                return
        elif os.path.isfile(args[0]):   #test for file existence
            self.file_name = args[0]
        else:
            logging.error('Error parsing function arguments.')

        if 'entries' in kwargs.keys():   #h5 file mode
            entries = kwargs['entries']
        else: entries = None

        if 'return_data' in kwargs.keys():
            return self.__acquire(entries=entries, return_data = kwargs['return_data'])
        else:
            self.__acquire(entries=entries, return_data = False)

    def __acquire(self, entries, return_data):
        self.errors = None
        self.optimized = False
        if self.file_name[-2:] == 'h5':
            self.discover_hdf_data(entries = entries)
            return self.read_hdf_data(return_data = return_data)
        else: #looks like a text file
            data = np.loadtxt(self.file_name, comments='#').T
            self.coordinate = self.coordinate
            self.data = data[self.cfg['data_column']]
            self.data_label = None
            if return_data: return data

    #\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=

    def read_hdf_data(self, return_data = False):
        '''
        Read hdf data and set attributes self.coordinate, self.data. 
        Return data as a numpy array if requested by 'return_data' argument.
        '''
        try:
            self.hf = store.Data(self.file_name)
        except (IOError,NameError):
            logging.error('Could not read h5 file.')
            return

        #read hdf data
        data = []
        for u in self.urls:
            data.append(np.array(self.hf[u],dtype=np.float64))
        self.hf.close()

        #fill coordinate and data attributes
        if len(self.urls) == 2:   #for only two urls, it is clear what to do
            self.coordinate = data[0]
            self.data = data[1]
            self.data_label = self.urls[1].split('/')[-1]
            self.data_url = self.urls[1]
        elif len(self.urls) > 2:   #if more than two, read out the data_column set in the config
            self.coordinate = data[0]
            self.data = data[self.cfg['data_column']]
            self.data_label = self.urls[self.cfg['data_column']].split('/')[-1]
            self.data_url = self.urls[self.cfg['data_column']]
        else:
            logging.warning('Coordinate and data attributes not assigned properly.')
        self.coordinate_label = self.urls[0].split('/')[-1]
        if return_data:
            return data, self.urls

    def find_latest_file(self, ftype):
        '''
        Find latest file of type 'ftype' in the data directory. 
        The file name (including absolute path) is stored in self.file_name.
        '''
     
        if self.cfg['datafolder_structure'] == 'day_time':
            self.data_dir = self.cfg['data_dir']
        else:
            self.data_dir = os.path.join(self.cfg['data_dir'], self.run_id, self.user)
        try:
            nfile = max(glob.iglob(str(self.data_dir)+'/*/*.'+ftype), key=os.path.getctime)   #find newest file in directory
        except ValueError:
            logging.error('No .{:s} file located in directory {:s}.'.format(str(ftype), str(self.data_dir)))
            return
        
        self.file_name = str(nfile).replace('\\','/')
        if self.cfg['show_output']: print('Latest file: {:s}'.format(self.file_name))

    def discover_hdf_data(self, entries=None):
        '''
        Read hdf data file and store urls that match entries or seem reasonable. 
        - Inputs: entries (optional): specifies entries in h5 file whose urls to be returned
        
        - entries can be a list of string keywords (entries) in the form ['string1',string2',...,'stringn']
        - when no entries are specified, discover_hdf_data looks for a frequency axis or a pulse length axis
          for the (first) coordinate axis and also searches amplitude_avg and phase_avg. If not present,
          the urls of 'amplitude' and 'phase' data are discovered

        - TODO: a nice feature would be to save a hint in each h5 file which entries to use for quick fitting
        '''
        
        try:
            self.hf = store.Data(self.file_name)
        except (IOError,NameError):
            logging.error('Could not read h5 file.')
            return
        
        keys = self.hf['/entry/data0'].keys()
        if self.cfg['show_output']: print('Available data entries:', keys)
        # only show the available data entries, but analysis entries can still be accessed
        
        url_tree = '/entry/data0/'
        urls = []
        
        if entries is None:   #no entries specified
            for k in keys:   #go through all keys
                try:
                    #search for parameter axis
                    if str(k[:4]).lower() == 'freq' or str(k[:2]).lower() == 'f ' or str(k[:4]).lower() == 'puls' or str(k[:4]).lower() == 'dacf' or str(k[:5]).lower() == 'delay' or str(k[:8  ]).lower() == 'pi pulse':
                        urls.append(url_tree + k)
                        break
                except IndexError:
                    logging.error('Entries cannot be identified. Parameter names too short. Aborting.')
                    return
            if len(urls) == 0:
                logging.error('No parameter axis found. Aborting.')
                return
                
            #first look for amplitude_avg and phase_avg entries -> time domain
            for k in keys:
                try:
                    if 'avg' in str(k).lower() and str(k[:3]).lower() == 'amp':
                        urls.append(url_tree + k)
                        break
                except IndexError:
                    logging.error('Entries cannot be identified. Aborting.')
                    return
                
            for k in keys:
                try:
                    if 'avg' in str(k).lower() and str(k[:3]).lower() == 'pha':
                        urls.append(url_tree + k)
                        break
                except IndexError:
                    logging.error('Entries cannot be identified. Aborting.')
                    return

            #if nothing found previously, then use amplitude and phase entries
            if len(urls) != 3:
                for k in keys:
                    try:
                        if str(k[:3]).lower() == 'amp':
                            urls.append(url_tree + k)
                            break
                    except IndexError:
                        logging.error('Entries cannot be identified. Aborting.')
                        return
                for k in keys:
                    try:
                        if str(k[:3]).lower() == 'pha':
                            urls.append(url_tree + k)
                            break
                    except IndexError:
                        logging.error('Entries cannot be identified. Aborting.')
                        return
            
        else:   #use specified entries
            entrytypes = self.hf['/entry/'].keys()
            try: entrytypes.remove('views')
            except (ValueError,AttributeError): pass
            for et in entrytypes:
                for e in entries:
                    for k in self.hf['/entry/'+et].keys():
                        try:
                            if e == et+k or (e == k and et == 'data0'):
                                urls.append('/entry/'+et+'/'+k)
                        except IndexError:
                            logging.error('Entries cannot be identified. No data for >> {:s} << found. Aborting.'.format(str(e)))
                            return

        self.hf.close()                
        #cast to real strings in case the urls ended up to be unicode
        self.urls = [str(u) for u in urls]
        if self.cfg['show_output']: print('Entries identified:', self.urls)

    #\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=

    def spline_smooth(self):
        '''
        Spline smooth the data set in self.data. The smoothing parameter is set via the 'spline_knots' entry in self.cfg.
        '''
        self.data = scipy.interpolate.UnivariateSpline(self.coordinate, self.data, s=self.cfg['spline_knots'])(self.coordinate)

    def phase_grad(self,spline = False):
        '''
        Unwrap, (optionally) spline smooth, and differentiate phase data. 
        The freuqnecy derivative of transmission or reflection phase data that follows 
        a arctan function yields a Lorentzian distribution.
        '''
        self.data = np.unwrap(self.data)
        if spline: self.spline_smooth()
        self.data = np.gradient(self.data)
        self._save_opt_data()

    #\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=

    def rotate_IQ_plane(self,entries=None):
        '''
        The data optimizer is fed with microwave data of both quadratures, typically amplitude and phase.
        It effectively performs a principle axis transformation in the complex plane of all data points
        that are located along a line for a typical projective qubit state measurement without single-shot
        fidelity and using a ADC acquisition card. Successively, data optimizer returns a projection along
        the in-line direction quadrature containing maximum information. Indeally, no information is lost
        as no data is left in the orthogonal quadrature.

        The algorithm locates one of the edge data points on either end of the line-shaped data point distribution
        by summing and maximizing the mutual distances between data points. It then calculates the distance
        of each data point with respect to this distinct extremal data point.

        We assume that the possible outcomes of a strong and projective quantum measurement, i.e. the qubit
        states |0>, |1>, result in two distinct locations in the complex plane, corresponding to signatures
        of the dispersive readout resonator. In the absence of single-shot readout fidelity, the pre-averaging
        that is perfomed by the ADC card in the complex plane leads to counts in between the piles denoting |0>
        and |1>. The points laong this line are distributed due to statistical measurement noise and ultimately
        due to the quantum uncertainty. As the averaging takes place in the complex plane, all points recorded
        in such a measurement should be located on a line. In case of a phase sensitive reflection measurement
        at a high internal quality cavity, points may span a segment of a circle that can be caused by noise
        caused measurements of the resonator position away from the two quantumn positions.

        The line shape of the data distribution generally justifies the projection of measurement data to an
        arbitrary axis without altering the internal data shape.

        Errors are calculated projected on the axis complex data points are aligned along. Data is normalized
        prior to returning.

        Requires coordinate, amplitude, and phase data to be specified via entries keyword. 
        self.amplitude and self.phase can be lists of vectors or a single (averaged) data set.
        '''

        if self.file_name[-2:] == 'h5':
            self.discover_hdf_data(entries = entries)
            if len(self.urls) != 3:
                logging.error('Invalid entry specification. Aborting.')
                return
            self.coordinate, self.amplitude, self.phase = self.read_hdf_data(return_data = True)[0]
        elif (self.amplitude is not None) and (self.phase is not None):
            pass
        else:
            logging.warning('Reading out amplitude and phase attributes.')

        #generate complex data array
        self.c_raw = np.array(self.amplitude) * np.exp(1j*np.array(self.phase))

        if len(self.c_raw.shape) > 1:
            #calculate mean of complex data
            c = np.mean(self.c_raw,axis=0)
        else:   #given data is already averaged
            c = self.c_raw

        #point in complex plane with maximum sumed mutual distances
        s = np.zeros_like(np.abs(c))
        for i in range(len(c)):
            for p in c:
                s[i]+=np.abs(p-c[i])
        cmax = np.extract(s == np.max(s),c)

        #calculate distances
        data_opt = np.abs(c - cmax)

        if len(self.c_raw.shape) > 1: #calculate errors in line direction
            #find maximum complex point in maximum distance
            d = 0
            for p in c:
                if np.abs(p-cmax) > d:
                    d = np.abs(p-cmax)
                    cdist = p
            #find unit vector in line direction
            vunit = (cdist - cmax)/np.abs(cdist - cmax)

            #calculate projected distances via dot product, projecting along the data direction
            #errors via std
            dist_proj = [0]*len(c)
            errs = np.zeros_like(c)
            for i,ci in enumerate(self.c_raw.T):   #for each iteration
                dist_proj[i] = [np.abs(np.vdot([np.real(vunit),np.imag(vunit)],[np.real(cr)-np.real(c[i]),np.imag(cr)-np.imag(c[i])])) for cr in ci]
                errs[i] = np.std(dist_proj[i])/np.sqrt(len(dist_proj[i]))
        
        #normalize optimized data
        data_opt -= np.min(data_opt)
        maxv = np.max(data_opt)
        data_opt /= maxv
        if len(self.c_raw.shape) > 1:
            errs /= maxv
        
        #gauss plane plot
        if self.cfg['show_complex_plot'] and self.cfg['matplotlib']:
            if len(self.c_raw.shape) > 1:
                plt.figure(figsize=(10,13))
                ax1 = plt.subplot2grid((4, 1), (0, 0))
                ax2 = plt.subplot2grid((4, 1), (1, 0), rowspan = 3)
                ax1.errorbar(data_opt,np.zeros_like(data_opt),xerr=np.real(errs),color='blue',fmt='o',elinewidth=0.8,capsize=5,markersize=8,ecolor='red')
                ax1.plot([0],[0],'*',color='red',markersize=20)
                prange = np.max(data_opt)-np.min(data_opt)
                ax1.set_xlim(np.min(data_opt)-0.05*prange,np.max(data_opt)+0.05*prange)
                
                ax2.plot(np.real(c),np.imag(c),'.')
                ax2.plot(np.real(c)[:10],np.imag(c)[:10],'.',color='r')   #show first 10 data points in red
                ax2.plot(np.real(cmax),np.imag(cmax),'*',color='black',markersize=15)
                self.errors = np.real(np.array(errs))
            else:
                plt.figure(figsize=(10,10))
                plt.plot(np.real(c),np.imag(c),'.')
                plt.plot(np.real(c)[:10],np.imag(c)[:10],'.',color='r')   #show first 10 data points in red
                plt.plot(np.real(cmax),np.imag(cmax),'*',color='black',markersize=15)

                self.errors = None
            
        self.data = np.real(np.array(data_opt))
        self._save_opt_data()

    def _save_opt_data(self):
        '''
        Saves optimized data in the h5 file and a respective view.
        '''
        if self.urls is not None:
            self.hf = store.Data(self.file_name)
    
            hdf_data_opt = self.hf.add_value_vector(self.data_label+'_data_opt', folder=self.cfg['analysis_folder'], x = self.hf.get_dataset(self.urls[0]))
            hdf_data_opt.append(np.array(self.data))
            self.optimized = True
            
            try:
                self.errors
                if self.errors is None: raise NameError
            except NameError: #no errors
                pass
            else:
                #write errors
                hdf_error = self.hf.add_value_vector(self.data_label+'_errors', folder=self.cfg['analysis_folder'])
                hdf_error.append(np.array(self.errors))
    
                #error plot view
                joint_error_view = self.hf.add_view(self.data_label+'_err_plot', x = self.hf.get_dataset(self.urls[0]),
                    y = hdf_data_opt, error = hdf_error)
            
            self.hf.close_file()

    #\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=
    
    def _store_fit_data(self, fit_params, fit_covariance):
        '''
        Appends fitted data to the h5 file in the specified analysis folder. As the fit is a fixed length array,
        a respective parameter axis is created and also stored in the h5 file.

        If data was optimized and stored with the method 'optimize', all rlevant joint views are created in the h5 file.
        
        inputs:
        - fit_params: np array of the resulting fit parameters
        - fit_covariance:  estimated covariances of the fit_params as returned by curve_fit
        '''
        
        try:
            self.hf = store.Data(self.file_name)
            
            #create coordinate and fit data vector
            hdf_x = self.hf.add_coordinate(self.data_label+'_coordinate',folder=self.cfg['analysis_folder'])
            hdf_x.add(self.x_vec)
            if self.optimized:
                hdf_y = self.hf.add_value_vector(self.data_label+'_opt_fit', folder=self.cfg['analysis_folder'], x = hdf_x)
            else:
                hdf_y = self.hf.add_value_vector(self.data_label+'_fit', folder=self.cfg['analysis_folder'], x = hdf_x)
            hdf_y.append(np.array(self.fvalues))
            
            #parameter entry including errors
            hdf_params = self.hf.add_coordinate(self.data_label+'_'+self.fit_function.__name__+'_params',folder=self.cfg['analysis_folder'])
            hdf_params.add(np.array(list(fit_params)+list(fit_covariance)))

            if self.data_label+'_data_opt' in self.hf[os.path.join('/entry/',self.cfg['analysis_folder']+'0').replace('\\','/')].keys():
                #create joint view with fit, data, and errors if existing
                try:
                    self.errors
                    if self.errors is None: raise NameError
                except NameError: #no errors
                    joint_view = self.hf.add_view(self.data_label+'_opt_fit', x = hdf_x, y = hdf_y)   #fit
                    joint_view.add(x = self.hf.get_dataset(self.urls[0]),
                        y = self.hf.get_dataset(os.path.join('/entry/', self.cfg['analysis_folder']+'0', self.data_label+'_data_opt').replace('\\','/')))   #data
                else: #inlcuding errors
                    joint_error_view_fit = self.hf.add_view(self.data_label+'_err_plot_fit', x = self.hf.get_dataset(self.urls[0]),
                        y = self.hf.get_dataset(os.path.join('/entry/', self.cfg['analysis_folder']+'0', self.data_label+'_data_opt').replace('\\','/')),
                        error = self.hf.get_dataset(os.path.join('/entry/', self.cfg['analysis_folder']+'0', self.data_label+'_errors').replace('\\','/')))   #errorplot
                    joint_error_view_fit.add(x = hdf_x, y = hdf_y)   #fit
            else:   #no optimization
                joint_view = self.hf.add_view(self.data_label+'_fit', x = hdf_x, y = hdf_y)   #fit
                joint_view.add(x = self.hf.get_dataset(self.urls[0]),
                    y = self.hf.get_dataset(self.data_url))   #data
            
            self.hf.close_file()
        except NameError as m:
            logging.error('Error while attempting to save fit data in h5 file: {:s}'.format(str(m)))

    #\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=

    def _guess_lorentzian_parameters(self):
        s_offs = np.mean(np.append(self.data[:int(0.1*len(self.data))],self.data[int(0.9*len(self.data)):]))
        #offset is calculated from the first and last 10% of the data to improve fitting on tight windows @andre20150318
        if np.abs(np.max(self.data) - np.mean(self.data)) > np.abs(np.min(self.data) - np.mean(self.data)):
            #expect a peak
            s_a = np.abs((np.max(self.data)-np.mean(self.data)))
            s_f0 = self.coordinate[np.where(self.data == max(self.data))[0][0]]*self.freq_conversion_factor
            if self.cfg['debug']:
                print('expecting peak')
                print(s_f0)
                print(s_a)
                print(s_offs)
        else:
            s_a = -np.abs((np.min(self.data)-np.mean(self.data)))
            s_f0 = self.coordinate[np.where(self.data == min(self.data))[0][0]]*self.freq_conversion_factor
            if self.cfg['debug']:
                print('expecting dip')
                print(s_f0)
                print(s_a)
                print(s_offs)
        
        #estimate peak/dip width
        mid = s_offs + 0.5*s_a   #estimated mid region between base line and peak/dip
        if self.cfg['debug']: print(mid)
        m = []   #mid points
        for dat_p in range(len(self.data)-1):
            if np.sign(self.data[dat_p] - mid) != np.sign(self.data[dat_p+1] - mid):   #mid level crossing
                m.append(dat_p)   #frequency of found mid point

        if len(m) > 1:
            s_k = (self.coordinate[m[-1]]-self.coordinate[m[0]])*self.freq_conversion_factor
            if self.cfg['show_output']: print('assume k = {:.3g}'.format(s_k))
        else:
            s_k = 0.15*(self.coordinate[-1]-self.coordinate[0])*self.freq_conversion_factor   #try 15% of window

        return [s_f0, s_k, s_a, s_offs]

    def _guess_oscillating_parameters(self,damping,asymmetric_exp = False):
        '''
        find initial parameters for oscillating fits
        
        data: data array
        data_c: (int) data column
        damping: (bool) switch for decaying/non-decaying fit functions
        asymmetric_exp
        '''
        
        #offset
        # testing last 25% of data for its maximum slope; take offset as last 10% of data for a small slope (and therefore strong damping)
        if np.max(np.abs(np.diff(self.data[int(0.75*len(self.data)):]))) < 0.3*np.abs(np.max(self.data)-np.min(self.data))/(len(self.coordinate[int(0.75*len(self.coordinate)):])*(self.coordinate[-1]-self.coordinate[int(0.75*len(self.data))])):   #if slope in last part small
            s_offs = np.mean(self.data[int(0.9*len(self.data)):])
        else:   #larger slope: calculate initial offset from min/max in data
            s_offs = (np.max(self.data) + np.min(self.data))/2
        #print s_offs
        
        if damping:
            if asymmetric_exp:
                s_a = np.abs(np.max(self.data) - np.min(self.data))
                if np.abs(np.max(self.data) - s_offs) < np.abs(np.min(self.data) - s_offs):   #if maximum closer to offset than minimum
                    s_a = -s_a
            else:
                a1 = np.abs(np.max(self.data) - s_offs)
                a2 = np.abs(np.min(self.data) - s_offs)
                s_a = np.max([a1,a2])
            #print s_a
            
            #damping
            a_end = np.abs(np.max(self.data[int(0.7*len(self.data)):]))   #scan last 30% of values -> final amplitude
            #print a_end
            # -> calculate Td
            t_end = self.coordinate[-1]
            try:
                s_Td = -t_end/(np.log((np.abs(a_end-np.abs(s_offs)))/s_a))
            except RuntimeWarning:
                if self.cfg['show_output']: logging.warning('Invalid value encountered in log. Continuing...')
                s_Td = float('inf')
            if np.abs(s_Td) == float('inf') and not asymmetric_exp:
                s_Td = float('inf')
                if self.cfg['show_output']: logging.warning('Consider using the sine fit routine for non-decaying sines.')
            if self.cfg['debug']: print('assume T = {:s}'.format(round(s_Td,4)))
        
        else:
            s_offs = np.mean(self.data)
            s_a = 0.5*np.abs(np.max(self.data) - np.min(self.data))
            s_Td = 0
        
        #frequency
        roots = 0   #number of offset line crossings ~ period of oscillation
        for dat_p in range(len(self.data)-1):
            if np.sign(self.data[dat_p] - s_offs) != np.sign(self.data[dat_p+1] - s_offs):   #offset line crossing
                roots+=1
        s_fs = float(roots)/(2*self.coordinate[-1])   #number of roots/2 /measurement time
        
        #phase offset calculation
        #the d are relative distances of the first data point
        dmax = np.abs(self.data[0] - np.max(self.data))
        dmean = np.abs(self.data[0] - np.mean(self.data))
        dmin = np.abs(self.data[0] - np.min(self.data))
        if dmax < dmean:   #start on upper side -> offset phase pi/2
            s_ph = np.pi/2
        elif dmin < dmean:   #start on lower side -> offset phase -pi/2
            s_ph = -np.pi/2
        else:   # sine or -sin?
            if np.sum(np.gradient(self.data)[:int(0.05*len(self.coordinate))]) > 0: #positive slope -> sin
                s_ph = 0
            else: #negative slope -> -sin
                s_ph = np.pi
        if self.cfg['debug']: print('assume phase = {:.3g}'.format(s_ph))
        
        if damping:
            return [s_fs, s_Td, s_a, s_offs, s_ph]
        else:
            return [s_fs, s_a, s_offs, s_ph]

    #\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=

    def fit_Lorentzian(self, p0=None):
        '''
        Regular Lorentzian fit.
        Optional Input: list of start parameters p0: Can be a smaller list than the actual number of parameters 
        (addressing only the first n start parameters). p0 can also contain 'None' entries in order to select 
        specific entries to set manually.
        '''
        self.fit_function = self.__f_Lorentzian
        self.fit(p0=p0)

    def fit_Skewed_Lorentzian(self, p0=None):
        self.fit_function = self.__f_Skewed_Lorentzian
        self.fit(p0=p0)
    
    def fit_Lorentzian_sqrt(self, p0=None):
        '''
        Square root of a Lorentzian fit e.g. for resonator magnitude data (since only the squared linear magnitude data 
        typically results in a Lorentzian response).
        Optional Input: list of start parameters p0: Can be a smaller list than the actual number of parameters 
        (addressing only the first n start parameters). p0 can also contain 'None' entries in order to select 
        specific entries to set manually.
        '''
        self.fit_function = self.__f_Lorentzian_sqrt
        self.fit(p0=p0)

    def fit_damped_sine(self, p0=None):
        '''
        Damped sine fit to be used e.g. to fit Rabi/Ramsey oscillations.
        Optional Input: list of start parameters p0: Can be a smaller list than the actual number of parameters 
        (addressing only the first n start parameters). p0 can also contain 'None' entries in order to select 
        specific entries to set manually.
        '''
        self.fit_function = self.__f_damped_sine
        self.fit(p0=p0)

    def fit_sine(self, p0=None):
        '''
        Sine fit to be used to fit non-decaying harmonic oscillations.
        Optional Input: list of start parameters p0: Can be a smaller list than the actual number of parameters 
        (addressing only the first n start parameters). p0 can also contain 'None' entries in order to select 
        specific entries to set manually.
        '''
        self.fit_function = self.__f_sine
        self.fit(p0=p0)

    def fit_exp(self, p0=None):
        '''
        Simple exponential fit.
        Optional Input: list of start parameters p0: Can be a smaller list than the actual number of parameters 
        (addressing only the first n start parameters). p0 can also contain 'None' entries in order to select 
        specific entries to set manually.
        '''
        self.fit_function = self.__f_exp
        self.fit(p0=p0)

    def fit_exp_sine(self, p0=None):
        '''
        Asymmetric exponential sine fit to be used to fit e.g. Vacuum Rabi oscillations.
        Optional Input: list of start parameters p0: Can be a smaller list than the actual number of parameters 
        (addressing only the first n start parameters). p0 can also contain 'None' entries in order to select 
        specific entries to set manually.
        '''
        self.fit_function = self.__f_exp_sine
        self.fit(p0=p0)

    #\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=\./=

    def fit(self,p0=None):
        '''
        Perform fitting and plotting based on the function stored in self.fit_function. 
        self.fit_function may be set manually with an arbitrary function before executing fit(self):

        .. code-block:: python
        
            def f_custom(coordinate, p1, p2, p3):
                return (coordinate+p1)*p2+np.exp(p3)
            self.fit_function = f_custom
            self.p0 = [1,2,3]   #must be provided
            self.fit()
            
        '''

        #check for unit in frequency
        if np.mean(self.coordinate) > 1.e6:
            self.freq_conversion_factor = 1e-9
            if self.cfg['show_output']: logging.warning('Frequency given in Hz. Conversion to GHz applied.')
        else:
            self.freq_conversion_factor = 1
    
        #start parameters
        if self.fit_function == self.__f_Lorentzian or self.fit_function == self.__f_Lorentzian_sqrt:
            self.guesses = self._guess_lorentzian_parameters()
        elif self.fit_function == self.__f_sine:
            self.guesses = self._guess_oscillating_parameters(damping=False,asymmetric_exp=False)
        elif self.fit_function == self.__f_damped_sine:
            self.guesses = self._guess_oscillating_parameters(damping=True,asymmetric_exp=False)
        elif self.fit_function == self.__f_exp_sine:
            self.guesses = self._guess_oscillating_parameters(damping=True,asymmetric_exp=True)
            d_diff = np.gradient(self.data,self.coordinate[1]-self.coordinate[0])  #true gradient
            i = 0
            for i in range(len(self.coordinate)):
                if np.sign(d_diff[i]) != np.sign(d_diff[i+1]):   #go to first sign change -> extremum
                    break
                i+=1
            if self.cfg['debug']: print('first extremum detected at {:.4g}'.format(self.coordinate[i]))
            s_offs = self.guesses[-1]
            s_d = 1 - np.abs(self.data[i] - s_offs)
            self.guesses += [s_d]
        elif self.fit_function == self.__f_exp:
            s_offs = np.mean(self.data[int(0.9*len(self.data)):])   #average over the last 10% of entries
            s_a = self.data[0] - s_offs
            #calculate gradient at t=0 which is equal to (+-)a/T
            s_Td = np.abs(float(s_a)/np.mean(np.gradient(self.data,self.coordinate[1]-self.coordinate[0])[:5]))
            self.guesses = [s_Td, s_a, s_offs]
        elif self.fit_function == self.__f_Skewed_Lorentzian:
            self.guesses = np.append(self._guess_lorentzian_parameters(), [1e-3, 1e-2]).tolist()
            self.p0 = _fill_p0(self.guesses, p0)
            try:
                self.popt, self.pcov = curve_fit(self.__f_Lorentzian, self.coordinate * self.freq_conversion_factor, self.data, p0=self.p0[:4], maxfev=10000)
            except Exception as e:
                logging.warning('Fit not successful.' + str(e))
                self.popt = self.p0[:4]
                self.pcov = None
            self.p0[:4] = self.popt
        else:
            self.guesses = None
            logging.warning('Fit function unknown. No guesses available but I will continue with the specified fit function.')
        self.p0 = _fill_p0(self.guesses,p0)

        try:
            self.popt, self.pcov = curve_fit(self.fit_function, self.coordinate*self.freq_conversion_factor, self.data, p0 = self.p0,maxfev=10000)
            if self.cfg['show_output'] and (self.fit_function == self.__f_Lorentzian or self.fit_function == self.__f_Lorentzian_sqrt):
                print('QL = {:.4g}'.format(np.abs(np.round(float(self.popt[0]) / self.popt[1]))))
        except:
            logging.warning('Fit not successful.')
            self.popt = self.p0
            self.pcov = None
        finally:
            self.x_vec = np.linspace(self.coordinate[0],self.coordinate[-1],500)

            self.fvalues = self.fit_function(self.x_vec * self.freq_conversion_factor, *self.popt)
            #plot
            create_plots = self.cfg['matplotlib'] and (
                self.cfg['show_plot'] or (
                    (self.cfg['save_png'] or self.cfg['save_pdf']) and self.file_name != "data_import"
                )
            )
            if create_plots:
                if self.fit_function == self.__f_exp:
                    #create pair of regular and logarithmic plot
                    self.fig, self.axes = plt.subplots(1, 2, figsize=(15,4))
                    
                    self.axes[0].plot(self.coordinate,self.data,'o')
                    self.axes[0].plot(self.x_vec, self.fvalues)
                    self.ax = self.axes[0]
                    
                    # log plot
                    self.axes[1].plot(self.coordinate,np.abs(self.data-self.popt[2]),'o')   #subtract offset for log plot
                    self.axes[1].plot(self.x_vec, np.abs(self.__f_exp(self.x_vec, *self.popt)-self.popt[2]))
                    self.axes[1].set_yscale('log')
                    self.axes[1].set_xlabel(self.coordinate_label, fontsize=13)
                    self.axes[1].set_ylabel('log('+self.data_label+')', fontsize=13)

                else:
                    #regular data plot
                    self.fig = plt.figure(figsize=self.cfg['figsize'])   #open and address plot instance
                    plt.plot(self.coordinate*self.freq_conversion_factor,self.data,'o')

                    plt.plot(self.x_vec*self.freq_conversion_factor, self.fvalues)
                    self.ax = plt.gca()

                self.ax.set_xlabel(self.coordinate_label, fontsize=13)
                self.ax.set_ylabel(self.data_label, fontsize=13)
                if self.guesses:
                    self.fig.suptitle(str(['{:s} = {:.4g}'.format(p, entry) for p, entry in zip(self.__get_parameters(self.fit_function), self.popt)]))
                    self.parameter_list = self.__get_parameters(self.fit_function)
                self.fig.tight_layout(rect=[0, 0, 1, 0.95])
                
                if self.cfg['save_png'] and self.file_name != "data_import": plt.savefig(self.file_name.strip('.h5')+'.png', dpi=self.cfg.get('dpi',200))
                if self.cfg['save_pdf'] and self.file_name != "data_import": plt.savefig(self.file_name.strip('.h5')+'.pdf', dpi=self.cfg.get('dpi',200))
                if self.cfg['show_plot']: plt.show()
                plt.close(self.fig)

        if self.pcov is not None and self.urls is not None: #if fit successful and data based on h5 file
            self._store_fit_data(fit_params=self.popt, fit_covariance=np.sqrt(np.diag(self.pcov)))
        if self.pcov is None:
            self.std = float('inf')*np.ones(len(self.popt)) #fill up errors with 'inf' in case fit did not converge
        else:
            self.std = np.sqrt(np.diag(self.pcov))



# ADwinProII_SMU.py driver for using ADwin Pro II as an SMU
# Author: Marius Frohn (uzrfo@student.kit.edu)
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import logging, time
from qkit.core.instrument_base import Instrument
import numpy as np
import ADwin

class ADwinProII_SMU(Instrument):
    """
    The ADwin Pro II by Jäger Computergesteuerte Messtechnik GmbH provides slots for various 
    ADC/DAC or data in/out cards as well as an accessable FPGA-like processor for control. Our
    group's device currently has a 4-channel 16-bit ADC, a 50kHz filtered 8-channel 16-bit ADC 
    and an 8-channel 16-bit DAC card. Getting the device running requires some steps outlined
    below:

    - Selecting drivers from https://www.adwin.de/de/download/download.html. Installing the 
      full software package is recommended for potential debugging, analyzing timings, etc.
    - 'pip install ADwin' to the qkit virtual environment. (Note: despite being present, the 
      driver could not correctly read out the ADwin install directory from the windows registry
      on my machine. To fix this, I commented out the entire try/except block around line ~100 
      in *venv*/site-packages/ADwin.py and hardcoded 'self.ADwindir = *adwin_install_path*')
    - The tool *adwin_install_path*/Tools/ADconfig/ADconfig.exe finds ADwin devices in the 
      network and allows e.g. assigning their MAC-addresses to a fixed IP-address. This should
      already be done for this device. In either case one still needs to register it as an 
      adwin-device on the measurement pc and assign it an adwin-device-number (default: 1)
    - With *adwin_install_path*/Tools/Test/ADpro/ADpro.exe one can check if communication with
      device is now possible & what the id-numbers of the used cards are (should be 1, 2, 3)
    - For usage one needs to provide a bootloader and to be run processes. The bootloader can 
      be found under *adwin_install_path*/ADwin12.btl for the T12 processor type. This driver 
      is in particular designed around 2 processes which configure the FPGA as a default SMU.
      If you're still reading this your problem can't be solved any other way but with this 
      box, so at this point I wish you good luck with whatever is troubling you. The already 
      compiled processes can be found as .TCx-files alongside the other files on the exchange 
      under exchange/Devices/ADwinProII/ and should be provided somewhere on the measurement pc. 

    SMU_SingleSetGet.TC1 enables ADC/DAC access by checking in regular intervals if a  
    respective action is requested. SMU_SweepLoop.TC2 is a high-priority process for sweeping 
    over a given array and reading back a given signal. The speed of one step is mostly limited 
    by the ADC/DAC access commands taking up to ~1600 CPU cycles = ~1.6e-6 s. The source code 
    for these processes is provided in the .bas-files. Alternatively one can program the FPGA 
    to do whatever one wants but then this driver obviously will no longer work with it. 
    """
    # static stuff
    ADC_CARD_OFFSET = 10
    ADC_FILTER_CARD_OFFSET = 20
    DAC_CARD_OFFSET = 30
    @staticmethod
    def _to_volt(x: np.ndarray) -> np.ndarray:
        return (x.astype(np.float64) - 0x8000) * 20.0 / 0x10000
    @staticmethod
    def _to_reg(x: np.ndarray) -> np.ndarray:
        x_reg = np.round((x + 10.0)/20.0*0x10000)
        x_reg = np.where(x_reg == -1, 0, x_reg)
        return np.where(x_reg == 0x10000, 0xFFFF, x_reg)
    
    def __init__(self, name: str, btl_path: str = "C:\\ADwin\\ADwin12.btl", proc1_path: str = "C:\\ADwin\\SMU_SingleSetGet.TC1", proc2_path: str = "C:\\ADwin\\SMU_SweepLoop.TC2", device_num: int = 1):
        """
        This is a driver for using the ADwinProII as an SMU by using the provided processes.

        Usage:
        Initialize with
        <name> = qkit.instruments.create('<name>', 'ADwinProII_SMU', **kwargs)

        Keyword arguments:
            btl_path:   path to bootloader file (default: "C:\\ADwin\\ADwin12.btl")
            proc1_path: path to process file containing low-priority ADC/DAC access
                        (default: "C:\\ADwin\\SMU_SingleSetGet.TC1")
            proc2_path: path to process file containing high-priority array sweep
                        (default: "C:\\ADwin\\SMU_SweepLoop.TC1")
            device_num: number assigned to device on your pc by ADconfig (default: 1)
        """
        # qkit stuff
        Instrument.__init__(self, name, tags=['physical'])
        self.add_parameter("adc", type = float, flags = Instrument.FLAG_GET, channels = (1, 4), minval = -10.0, maxval=10.0, units = "V")
        self.add_parameter("adc_filtered", type = float, flags = Instrument.FLAG_GET, channels = (1, 8), minval = -10.0, maxval=10.0, units = "V")
        self.add_parameter("dac", type = float, flags = Instrument.FLAG_SET, channels = (1, 8), minval = -10.0, maxval = 10.0, units = "V")
        # Boot & Load processes
        self.adw = ADwin.ADwin(device_num)
        self.adw.Boot(btl_path)
        logging.info("Booted ADwinProII with processor {}".format(self.adw.Processor_Type()))
        self.adw.Load_Process(proc1_path)
        self.adw.Start_Process(1)
        logging.info("Loaded SMU process 1 for ADwinProII's FPGA")
        self.adw.Load_Process(proc2_path)
        logging.info("Loaded SMU process 2 for ADwinProII's FPGA")
        # Sweep parameters
        self.dac_channel = 1 # 1...8
        self.adc_channel = 1 # 1...4 or 1...8 depending on card
        self.adc_card = 1 # 1 (normal), 2 (filtered)
        self._sweep_channels = (self.dac_channel, self.adc_channel) # for qkit compability
        self.delay = 2000 # NOTE: int describing to be slept time inbetween dac set and adc get. 
        # slept time <=> self.delay * 1e-8 s, set/get commands take ~2e-6 s per point on their own always

    # functionality
    def do_get_adc(self, channel: int) -> float:
        """
        This is a very nice docstring
        """
        if channel in range(1, 5):
            self.adw.Set_Par(channel + self.ADC_CARD_OFFSET, 1)
            for i in range(5000):
                if self.adw.Get_Par(channel + self.ADC_CARD_OFFSET) == 0:
                    return self.adw.Get_FPar(channel + self.ADC_CARD_OFFSET)
                time.sleep(0.001)
            raise TimeoutError("ADwin Pro II did not confirm ADC read within 5s")
        else:
            raise ValueError("Channel must be 1...4")
        
    def do_get_adc_filtered(self, channel: int) -> float:
        """
        This is a very nice docstring
        """
        if channel in range(1, 9):
            self.adw.Set_Par(channel + self.ADC_FILTER_CARD_OFFSET, 1)
            for i in range(5000):
                if self.adw.Get_Par(channel + self.ADC_FILTER_CARD_OFFSET) == 0:
                    return self.adw.Get_FPar(channel + self.ADC_FILTER_CARD_OFFSET)
                time.sleep(0.001)
            raise TimeoutError("ADwin Pro II did not confirm ADC read within 5s")
        else:
            raise ValueError("Channel must be 1...8")
        
    def do_set_dac(self, volt: float, channel: int) -> None:
        """
        This is a very nice docstring
        """
        if channel in range(1, 9):
            self.adw.Set_FPar(channel + self.DAC_CARD_OFFSET, volt)
            self.adw.Set_Par(channel + self.DAC_CARD_OFFSET, 1)
        else:
            raise ValueError("Channel must be 1...8")

    def set_sweep_parameters(self, sweep: np.ndarray) -> None:
        """
        Check to be swept parameter settings and write them to the device.
        Sweep: array describing sweep values by [start, stop, step]

        Max 2^16 points due to (arbitary) memory limitation on device (More 
        values would not make sense anyways though since the ADC/DACs have 
        only 16-bit resolution)
        Voltage range is limited to +-10V for both ADC/DACs
        Look at device for channel number limitations
        """
        # self.dac_channel, self.adc_channel = self._sweep_channels # NOTE: keep this line? potentially causes more problems with qkit defaults not fitting for this device than the feature is worth 
        set_array = np.arange(sweep[0], sweep[1] + sweep[2]/2, sweep[2])
        # Check values
        if len(set_array) > 0x1000:
            raise ValueError("Sweep size of 16 bit ADC/DACs limited to 2^16 values")
        if np.any(np.abs(set_array) > 10.0):
            raise ValueError("Sweep array contains values outside +-10V range")
        if not self.dac_channel in range(1, 9):
            raise ValueError("DAC channel must be 1...8")
        if not self.adc_card in [1, 2]:
            raise ValueError("ADC card must be 1 (normal), 2 (50kHz filtered)")
        if self.adc_card == 1 and not self.adc_channel in range(1, 5):
            raise ValueError("ADC channel must be 1...4")
        if self.adc_card == 2 and not self.adc_channel in range(1, 9):
            raise ValueError("Filtered ADC channel must be 1...8")
        # Set values
        self.adw.Set_Par(1, len(set_array))
        self.adw.Set_Par(2, self.dac_channel)
        self.adw.Set_Par(3, self.adc_channel)
        self.adw.Set_Par(4, self.adc_card)
        self.adw.Set_Par(5, int(self.delay))
        self.adw.SetData_Long(self._to_reg(set_array), 1, 1, len(set_array))

    def get_tracedata(self) -> tuple[np.ndarray]:
        """
        Starts a sweep with parameters currently set on device and returns
        set_values_array, measured_values_array. Sorting by what is I and V
        is being handled by overlaying virtual_tunnel_electronic
        """
        # Sweep
        self.adw.Start_Process(2)
        while self.adw.Process_Status(2):
            time.sleep(0.1)
        # Read result
        return self._to_volt(self.adw.GetData_Long(1, 1, self.adw.Get_Par(1))), self._to_volt(self.adw.GetData_Long(2, 1, self.adw.Get_Par(1)))
    
    # qkit SMU compability
    def set_sweep_mode(self, mode: int = 0):
        if mode != 0:
            print("ADwinProII only has voltage ADC/DACs, only VV-mode 0 supported")
    def get_sweep_mode(self) -> int:
        return 0
    def get_sweep_channels(self) -> tuple[int]:
        return (self.dac_channel, self.adc_channel)
    def set_status(self, *args, **kwargs) -> None:
        pass # ADC/DACs are always responsive
    def reset(self):
        for i in range(1, 9):
            self.do_set_dac(0.0, i)



import qkit
import numpy as np
import os.path
import time
import logging,warnings
import sys, gc
#from gui.notebook.Progress_Bar import Progress_Bar
import qkit.measure.timedomain.awg.load_awg as load_awg
import qkit.measure.timedomain.awg.generate_waveform as gwf



iq = qkit.instruments.get('iq')
gc.collect()

def concatenate_pulses(pulse_durations,phases, positions,sample):
    '''
        generate waveform with varoius pulses

        Input:
            pulse_durations - 1d array of pulse lengths (in s)
            phases - 1d array of phases of the pulses (in rad)
            positions - 1d array of the pulse position, measured in s from the end of the pulse to the end of the wfm (smallest distance)
        NOTE that all input arrays have to be of the same length!
        
            length - length of the wfm (in s), also known as exc_T
        
        Output:
            (complex) float array of samples
    '''
    if not(len(pulse_durations)==len(phases) and len(phases)==len(positions)):
        raise ValueError("Input Arrays do not have the same size: pulse_durations: %i, phases: %i, positions: %i"%
            (len(pulse_durations),len(phases),len(positions)))
    
    #if(pulses*(delay+pi2_x_pulse)-delay > length): logging.error(__name__ + ' : x-pulses do not fit into waveform')
    
    phases=np.exp(1j*np.array(phases))

    wfm = np.zeros_like(gwf.square(0, sample),dtype=np.complex)    

    for i in range(len(pulse_durations)):
        wfm += gwf.square(pulse_durations[i], sample, position= length-positions[i])*phases[i]
    return wfm
	
def append_wfm(large_wfm, appendix):
	return np.append(large_wfm[len(appendix):],appendix)
	

def length(thetas, phis, wfm = None, sample = None):
	'''
		returns the number of tomography steps.
		You will need this for your measurement.
	'''
	angles=[[0,0]]
	for i,th in enumerate(thetas):
		angles=np.append(angles,np.array([np.ones(phis[i])*th,np.linspace(0,2*np.pi,phis[i],endpoint=False)]).T,axis=0)
	angles=angles[1:]
	return len(angles)
	
def radial(thetas, phis, wfm, sample, marker = None, delay = 0, markerfunc = None):
	angles=[[0,0]]
	for i,th in enumerate(thetas):
		angles=np.append(angles,np.array([np.ones(phis[i])*th,np.linspace(0,2*np.pi,phis[i],endpoint=False)]).T,axis=0)
	angles=angles[1:]
	if not os.path.exists(qkit.cfg.get('datadir')+time.strftime("\\%Y%m%d")):
		os.makedirs(qkit.cfg.get('datadir')+time.strftime("\\%Y%m%d"))
	np.savetxt(qkit.cfg.get('datadir')+time.strftime("\\%Y%m%d\\Tomography_%H%M%S.set"),angles)
	sample.update_instruments()
	if marker == None:
		new_marker = None
	else:
		new_marker = [[],[],[],[]]
		if marker.shape[0] == 4: #each marker is defined
			for i in range(len(marker)):
				new_marker[i] = [ append_wfm(marker[i],np.zeros_like(gwf.square(0, sample, angle[0]/np.pi*sample.tpi + delay))) for angle in angles ]
		else:
			new_marker[0] = [ append_wfm(marker,np.zeros_like(gwf.square(0, sample,  angle[0]/np.pi*sample.tpi + delay))) for angle in angles ]
			new_marker[1:4]=[ np.zeros_like(new_marker[0]) for i in range(3) ]
		new_marker = [[new_marker[0],new_marker[1]],[new_marker[2],new_marker[3]]]
	result = load_awg.update_sequence(range(len(angles)), lambda t, sample2: iq.convert(append_wfm(wfm,gwf.square(angles[t][0]/np.pi*sample.tpi, sample, angles[t][0]/np.pi*sample.tpi + delay)*np.exp(1j*angles[t][1]))), sample, marker = new_marker, markerfunc=markerfunc)
	print "You have a tomography resolution of %i points"%len(angles)
	return result

### Use like this:
# thetas = np.linspace(0,2*np.pi, 20)
# phis = [1]+(len(thetas)-1)*[15] #for 15 steps in phi direction (but 0 rotation has only 1 pulse)
# gt.radial(thetas, phis, gwf.square(qubit.tpi, qubit), qubit)

def threepoint(ts, wfm_func, sample, loop = False, drive = 'c:', path = '\\waveforms', reset = True, marker=None, markerfunc = None, largeblocks = False, delay = 0):
	'''
		This is a helper function to pass the right sequences to your AWG for a tomographic measurement.
		
		For each value in ts, three measurements are done:
		* One without a rotation
		* One with a rotation around x just before the readout
		* One with a rotation around y just before the readout
		If largeblocks=False (default): all 3 mmts are done for one value in ts, then go on to the next value. If largeblocks=True, the mmts consists of 3 large blocks with all ts, so the tomographic mmt sequence changes slowly.
		
		The wfm_func is automatically shifted by a pi/2 pulse + delay, THE MARKER IS ONLY SHIFTED WHEN PASSED AS markerfunc!
		
		
		
		At the end, three mmts are appended: no pulse, pi/2 pulse and pi-pulse, all without any tomographic pulse, in order to get a good scale for the analysis.
	'''
	
	def phase_t(t): 
		if largeblocks: #In this version, all ts will be gone through first and then all phases
			phase = t/len(ts)
			t = t%len(ts)
			if phase == 3 :
				return [-1-t,0]
			return [ts[t],[0,1,1j][phase]]
			phase = [0,1,1j][phase]  #map the index to the right values
			t = ts[t]
			return [t,phase]

		else:	#In this version, all phases will be gone through first and then all ts
			phase = t%3
			t = t/3
			if t >= len(ts) :
				return [-1-phase,0]
			
			return [ts[t],[0,1,1j][phase]]
			phase = [0,1,1j][phase]  #map the index to the right values
			return [t,phase]
	
	def wfm_helper (t, sample):
		t,phase = phase_t(t)
		if t == -1 : return gwf.square(0,sample)
		if t == -2 : return gwf.square(sample.tpi,sample)
		if t == -3 : return gwf.square(sample.tpi2,sample)
		return append_wfm(wfm_func(t,sample),gwf.square(sample.tpi2,sample,sample.tpi2+delay)*phase)
	
	def marker_helper (t, sample):
		t,phase = phase_t(t)
		if t <0 : return gwf.square(0,sample)
		return append_wfm(markerfunc(t,sample),gwf.square(0,sample,sample.tpi2+delay)) #ToDo: This does not work for the general array markerfunc=[[funcA1,None],[None,funcB2]]
	
	ts2 = range(3*len(ts)+3)
	if marker!= None:
		if largeblocks:		#The following is for option 1, where first all x, then all y coordinates are taken
			marker = np.append(
						np.append(
							np.append(
								marker,
								marker,axis=2),
							marker,axis=2), # Add the same marker three times because of the 
						np.zeros_like(marker[:,:,:3,:]),axis=2)
			
		else:	#The folllowing is for option 2, where all coordinates are taken for one ts:
			marker = np.append(marker.repeat(3,axis=2),np.zeros_like(marker[:,:,:3,:]),axis=2)
	if markerfunc==None:
		return load_awg.update_sequence(ts2, lambda t, sample: iq.convert(wfm_helper(t,sample)), sample, loop = loop, drive = drive, path = path, reset = reset, marker=marker ,markerfunc=None)
	else:
		return load_awg.update_sequence(ts2, lambda t, sample: iq.convert(wfm_helper(t,sample)), sample, loop = loop, drive = drive, path = path, reset = reset, marker=marker ,markerfunc=lambda t, sample:marker_helper(t,sample))


#!/usr/bin/env python

# agilent_33210A class, derived from Picowatt_AVS47 class by Hannes Rotzinger.
# Jochen Zimmer, June 2012 -- ALPHA release version 0.1
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA


QTLAB = True

if QTLAB:
    from qkit.core.instrument_base import Instrument
else:
    class Instrument(object):
        FLAG_GETSET = 0 
        FLAG_GET_AFTER_SET = 1
        def __init__(self,name,**kwargs):
            pass
        def add_parameter(self,name,**kwargs):
            pass
        def add_function(self,name):
            pass

from qkit import visa
from time import sleep
import sys

class agilent_33210A(Instrument):

    def __init__(self, name, address):

        Instrument.__init__(self, name, tags=['physical'])

        self._address = address
        self._visainstrument = visa.instrument(self._address)

        self.min_freq = self.get_MIN_frequency()
        self.max_freq = self.get_MAX_frequency()

        self.min_ampl = self.get_MIN_amplitude()
        self.max_ampl = self.get_MAX_amplitude()

        self.min_offs = self.get_MIN_offset()
        self.max_offs = self.get_MAX_offset()

        self.set_default() #reset

        self.add_function('set_default')
        self.add_function('get_frequency')
        self.add_function('set_frequency')
        self.add_function('get_function')
        self.add_function('set_function')
        self.add_function('get_amplitude')
        self.add_function('set_amplitude')
        self.add_function('get_output')
        self.add_function('set_output')
        self.add_function('get_offset')
        self.add_function('set_offset')
        self.add_function('get_voltage_units')
        self.add_function('set_voltage_units')
        self.add_function('get_output_load')
        self.add_function('set_output_load')


    def set_default(self):
        com_str = "*RST"
        self._visainstrument.write(com_str)
        sleep(0.1)
        return 0

    def get_frequency(self):
        com_str = "FREQuency?"
        self._visainstrument.write(com_str)
        res_str = self._visainstrument.read()
        return float(res_str)

    def get_MAX_frequency(self):
        com_str = "FREQuency? MAX"
        self._visainstrument.write(com_str)
        res_str = self._visainstrument.read()
        return float(res_str)

    def get_MIN_frequency(self):
        com_str = "FREQuency? MIN"
        self._visainstrument.write(com_str)
        res_str = self._visainstrument.read()
        return float(res_str)

    def set_frequency(self, freq):
        freq = float(freq)
        if (freq<self.min_freq) or (freq>self.max_freq):
            print("agilent_33210A: frequency out of range!", file=sys.stderr)
            return -1
        com_str = "FREQuency %f" % freq
        self._visainstrument.write(com_str)
        sleep(0.05)
        return self.get_frequency()

    def get_function(self):
        com_str = "FUNCtion?"
        self._visainstrument.write(com_str)
        res_str = self._visainstrument.read()
        return res_str

    def set_function(self,f_str):
        if (f_str != "SIN") and (f_str != "SQU"):
            print("agilent_33210A: selected function not recognized!", file=sys.stderr)
            # FUNCtion {SINusoid|SQUare|RAMP|PULSe|NOISe|DC|USER}, some not implemented
            return -1
        com_str = "FUNCtion %s" % f_str
        self._visainstrument.write(com_str)
        sleep(0.05)
        return self.get_function()

    def get_amplitude(self):
        com_str = "VOLTage?"
        self._visainstrument.write(com_str)
        res_str = self._visainstrument.read()
        return float(res_str)

    def get_MAX_amplitude(self):
        com_str = "VOLTage? MAX"
        self._visainstrument.write(com_str)
        res_str = self._visainstrument.read()
        return float(res_str)

    def get_MIN_amplitude(self):
        com_str = "VOLTage? MIN"
        self._visainstrument.write(com_str)
        res_str = self._visainstrument.read()
        return float(res_str)

    def set_amplitude(self, ampl):
        ampl = float(ampl)
        if (ampl<self.min_ampl) or (ampl>self.max_ampl):
            print("agilent_33210A: amplitude out of range!", file=sys.stderr)
            return -1
        com_str = "VOLTage %f" % ampl
        self._visainstrument.write(com_str)
        sleep(0.05)
        return self.get_amplitude()

    def get_output(self):
        com_str = "OUTPut?"
        self._visainstrument.write(com_str)
        res_str = self._visainstrument.read()
        return res_str

    def set_output(self, onoff = 0):
        onoff = int(bool(onoff))
        com_str = "OUTPut %i" % onoff
        self._visainstrument.write(com_str)
        sleep(0.05)
        return self.get_output()

    def get_offset(self):
        com_str = "VOLTage:OFFset?"
        self._visainstrument.write(com_str)
        res_str = self._visainstrument.read()
        return float(res_str)

    def get_MAX_offset(self):
        com_str = "VOLTage:OFFset? MAX"
        self._visainstrument.write(com_str)
        res_str = self._visainstrument.read()
        return float(res_str)

    def get_MIN_offset(self):
        com_str = "VOLTage:OFFset? MIN"
        self._visainstrument.write(com_str)
        res_str = self._visainstrument.read()
        return float(res_str)

    def set_offset(self, offs = 0):
        offs = float(offs)
        if (offs > self.max_offs) or (offs < self.min_offs):
            print("agilent_33210A: offset out of range!", file=sys.stderr)
            return -1
        com_str = "VOLTage:OFFset %f" % offs
        self._visainstrument.write(com_str)
        sleep(0.05)
        return self.get_offset()

    def get_voltage_units(self):
        com_str = "VOLTage:UNIT?"
        self._visainstrument.write(com_str)
        res_str = self._visainstrument.read()
        return res_str

    def set_voltage_units(self, vunits = "VPP"):
        if (vunits != "VPP") and (vunits != "VRMS") and (vunits != "DBM"):
            print("agilent_33210A: input error in set_voltage_units()", file=sys.stderr)
            return -1
        com_str = "VOLTage:UNIT %s" % vunits
        self._visainstrument.write(com_str)
        sleep(0.05)
        self.min_ampl = self.get_MIN_amplitude() # depends on the units, of course
        self.max_ampl = self.get_MAX_amplitude()
        return self.get_voltage_units()

    def get_output_load(self):
        com_str = "OUTPut:LOAD?"
        self._visainstrument.write(com_str)
        res_str = self._visainstrument.read()
        return float(res_str)

    def set_output_load(self, load):
        if (load == "INF"):
            com_str = "OUTPut:LOAD %s" % load
        elif float(load) > 1.0:
            com_str = "OUTPut:LOAD %f" % float(load)
        else:
            print("agilent_33210A: input error in set_output_load()", file=sys.stderr)
            return -1
        self._visainstrument.write(com_str)
        sleep(0.05)
        self.min_ampl = self.get_MIN_amplitude() # also depends on load
        self.max_ampl = self.get_MAX_amplitude()
        return self.get_output_load()


## testing:

if __name__ == "__main__":

    agilent_AWG = agilent_33210A("agilent_AWG", "TCPIP::192.168.0.2")

    print( agilent_AWG._visainstrument.ask("*IDN?"))
    
    print( agilent_AWG.set_voltage_units("VRMS"))
        
    print( agilent_AWG.set_output_load("INF"))

    print( agilent_AWG.set_frequency(1000.0))

    print( agilent_AWG.set_amplitude(0.1))

    print( agilent_AWG.set_offset(0.1))
    
    print( agilent_AWG.set_output(1))

    sleep(1)

    print( agilent_AWG.set_output(0))

#    start = float(clock())
#    for i in range(100):
#        hp_mm.get_configuration()
#    end = float(clock())    
#    print "Getting conf took:", end-start






import smbus
import logging


class IOPort(list):
    """
    Represents the PCF8575 IO port as a list of boolean values.
    """
    def __init__(self, pcf8575, *args, **kwargs):
        super(IOPort, self).__init__(*args, **kwargs)
        self.pcf8575 = pcf8575

    def __setitem__(self, key, value):
        """
        Sets an individual output pin to value.
        """
        self.pcf8575.set_output(key, value)
        logging.info(f'set port {key} to {value}')

    def __getitem__(self, key):
        """
        Gets an individual pin state.
        """
        return self.pcf8575.get_output(key)

    def __repr__(self):
        """
        Represent all ports as a list of booleans.
        """
        state = self.pcf8575.bus.read_word_data(self.pcf8575.address, 0)
        ret = []
        for i in range(16):
            ret.append(bool(state & 1 << 15 - i))
        return repr(ret)

    def __len__(self):
        return 16

    def __iter__(self):
        for i in range(16):
            yield self[i]

    def __reversed__(self):
        for i in range(16):
            yield self[15 - i]


class PCF8575(object):
    """
    A software representation of a single PCF8575 IO expander chip.
    """
    def __init__(self, i2c_bus_no, address):
        self.bus_no = i2c_bus_no
        self.bus = smbus.SMBus(i2c_bus_no)
        self.address = address

    def __repr__(self):
        return f'PCF8575(i2c_bus_no={self.bus_no:d}, address={self.address:#02x})'

    @property
    def port(self):
        """
        Represent IO ports as a list of boolean values.
        """
        return IOPort(self)

    @port.setter
    def port(self, value):
        """
        Set all ports using a list.
        """
        assert isinstance(value, list)
        assert len(value) == 16
        new_state = 0
        for i, val in enumerate(value):
            if val:
                new_state |= 1 << 15 - i
        self.bus.write_byte_data(self.address, new_state & 0xff, (new_state >> 8) & 0xff)

    def set_output(self, pin, value):
        """
        Sets a specific pin output to high (True) or low (False).
        """
        assert pin in range(16), "Output number must be an integer between 0 and 15"
        current_state = self.bus.read_word_data(self.address, 0)
        bit = 1 << 15 - pin
        new_state = current_state | bit if value else current_state & (~bit & 0xffff)
        self.bus.write_byte_data(self.address, new_state & 0xff, (new_state >> 8) & 0xff)

    def get_output(self, pin):
        """
        Gets the boolean state of an individual pin.
        """
        assert pin in range(16), "Pin number must be an integer between 0 and 15"
        state = self.bus.read_word_data(self.address, 0)
        return bool(state & 1 << 15 - pin)



import qkit
from qkit.core.instrument_base import Instrument
import numpy as np
from time import sleep

class virtual_magnet(Instrument):

    def __init__(self, name, source=None, channel = 0, stepsize = 2e-9, stepdelay=1, multiplier = 1, range = 1e-3, old = 0):
        Instrument.__init__(self, name, tags=['virtual'])
        self._instruments = qkit.instruments.get_instruments()
        self._source= self._instruments.get(source)
        self._channel = channel        
        # Add parameters        
        self.add_parameter('current', type=float,
            flags=Instrument.FLAG_SET, units='A')
        self.add_parameter('stepsize', type=float,
            flags=Instrument.FLAG_GETSET, units='A')        
        self.add_parameter('stepdelay', type=float,
            flags=Instrument.FLAG_GETSET, units='ms')
        self.add_parameter('multiplier', type=float,
            flags=Instrument.FLAG_GETSET, units='')
        self.add_parameter('range', type=float,
            flags=Instrument.FLAG_GETSET, units='A')            

        self._stepsize = stepsize
        self._stepdelay = stepdelay
        self._multiplier = multiplier
        self._range = range
        self._old = old;
        self.get_all()
    def get_all(self):
        self.get_stepsize()
        self.get_stepdelay()
        self.get_range()
        self.get_multiplier()

    def do_set_current( self, cur):
                if np.abs(cur) > 10*self._range:
                        raise ValueError('set_current(): given value exeeds the range of %f' % self._range)
        old = self._old
        stepsize =  self.get_stepsize()
        stepdelay = self.get_stepdelay()
        f = 1/(self._multiplier*self._range) #conversion factor from current to voltage
        if np.abs(old-cur) <= stepsize:
            steps = cur*f
        elif old > cur:                
            steps = np.arange(old*f,cur*f-stepsize*f,-stepsize*f)
        else:
            steps = np.arange(old*f,cur*f+stepsize*f,stepsize*f)
        if np.size(steps)>1:    
            for i in steps:
                self._source.set('ao%i' % self._channel, i )
                sleep(stepdelay*1e-3)
        self._source.set('ao%i' % self._channel,cur*f)    
        self._old = cur #Putting new value to the buffer        
        
    def do_get_stepsize(self):
        return self._stepsize
    def do_set_stepsize(self, val):
        self._stepsize = val
        
    def do_get_stepdelay(self):
        return self._stepdelay
    def do_set_stepdelay(self, val):
        self._stepdelay = val
        
    def do_get_range(self):
        return self._range
    def do_set_range(self, val):
        self._range = val     

    def do_get_multiplier(self):
        return self._multiplier
    def do_set_multiplier(self, val):
        self._multiplier = val              



# -*- coding: utf-8 -*-
"""
keeping track of your measurement files 

@author: MP, AS, TW, HR@KIT / 2018
@license GPL


file info database
==================


lets introduce a set of 
fid configuration settings (defaults)
=====================================
fid_scan_datadir = True
    Indicates whether you want to scan your datadir for h5 files at startup.
fid_scan_hdf     = False
    This will open every h5 file found and extract attributes.
fid_init_viewer  = True
    Make a database out of the dictionary of h5 files.


databases
=========

h5_db
----------
This is the main database. It holds an index of UUID <-> h5-file path.
usage: 
qkit.fid.h5_db.get("UUID") or qkit.fid.get("UUID") returns the h5_file path

h5_info_db
----------
This database is only populated when the qkit.cfg setting 'fid_scan_hdf' is True.
The database holds an UUID index with extended informations about the hdf file. 
populating this database is much slower than h5_db therfore not on by default.

set_db and measure_db
------
Hold an UUID index with the settings for a h5_file


basic usage
===========

this is done automatically on qkit startup: (in n.op.)
-------------------------------------------
import qkit.core.lib.file_service.file_info_database as fid
qkit.fid = fid.fid()

qkit.view(file_id)
-----------
opens a qviewkit instance with the file_id


import and basic usage of the h5-grid-viewer
============================================
# qgrid is an interactive jupyter tool for pandas dataframe, which helps filtering your data
# for more information see https://github.com/quantopian/qgrid
# script works also well without qgrid
# you can simply access all the data via pandas commands, i.e., qkit.fid.df.columnname['uid']
# for further information see doc strings


after startup, the following main commands are available in a jupyter notebook:
-------------------------------------------------------------------------------
qkit.fid.show()
qkit.fid.view(file_id)

qkit.fid.get_uuid(time)
qkit.fid.get_time(uuid)
qkit.fid.get_date(uuid)
"""



import logging
import threading
from distutils.version import LooseVersion

import numpy as np

import qkit

found_qgrid = False
if qkit.module_available['pandas']:
    import pandas as pd
    if qkit.module_available['qgrid']:
        import qgrid as qd
        found_qgrid = True

from qkit.core.lib.file_service.file_info_database_lib import file_system_service

# display using qviewkit
from qkit.gui.plot.plot import plot

class fid(file_system_service):
    def __init__(self):
        self._batch_update = False
        self.column_sorting = ['datetime', 'name', 'run', 'user', 'comment', 'rating']
        self.columns_ignore = ['time']
        self.df = None
        # create initial database in the background. This can take a while...
        self.create_database()
        self._selected_df = []
        self.found_qgrid = found_qgrid
        

    history = property(lambda self: sorted(self.h5_db.keys()))
    
    def get_last(self):
        return sorted(self.h5_db.keys())[-1]
        
    def __getitem__(self, key):
        with self.lock:
            if type(key) == int:
                return sorted(self.h5_db.keys())[key]
            try:
                return self.h5_db[key]
            except KeyError as e:
                raise KeyError("Can not find your UUID '{}' in qkit.fid database.".format(key))

    def get(self, key, args=None):
        with self.lock:
            if key not in self.h5_db:
                logging.error("Can not find your UUID '{}' in qkit.fid database.".format(key))
            return self.h5_db.get(key, args)

    def view(self, file_id=None):
        """
        view a datafile with qviewkit
        =============================
    
        Args:
           file_id (str):  identifier for a data file.
           
           file_id can be 
           
           None (empty): then the latest file is opened
           
           int: e.g. -1, 100 
                then the number is opened from the last list
           list: e.g ['UUID1' , 'UUID2']
               each UUID is opened
           uuid:
               the data file is looked up via the h5_db and opened
           path:
               the data file is opened using the path 

        Returns:
            None
        """

        def plotif(filename):
            if filename:
                plot(filename, live=False)

        if qkit.module_available['pandas']:
            if type(file_id) is pd.Index:
                file_id = list(file_id)
        
        if file_id is None:
            plotif(self.h5_db.get(self.get_last()))
        elif type(file_id) is int:
            plotif(self.h5_db.get(self.history[file_id]))
        elif type(file_id) is list:
            for i in file_id:
                plotif(self.h5_db.get(i, False))
        elif type(file_id) is str:
            plotif(self.h5_db.get(file_id, False))

    def create_database(self,block=False):
        t1 = threading.Thread(name='creating_db', target=self.update_all)
        t1.start()
        if block:
            t1.join()
    
    def recreate_database(self):
        '''
            Deletes all cached database files and rescans the whole directory tree.
            Use this if your database looks strange.
        '''
        self.h5_db = {}
        self.set_db = {}
        self.measure_db = {}
        self.h5_info_db = {}
        self._h5_n_mtime = {}
        
        self._remove_cache_files()
        self.create_database()

    def update_all(self):
        """ updates file and grid database if activated
        """
        self.update_file_db()
        self.update_grid_db()

    def update_grid_db(self):
        with self.lock:
            if qkit.cfg.get('fid_init_viewer',qkit.module_available['pandas']):  # Pandas ist needed here, so if fid_init_viewer is not set, we make it dependent on pandas
                qkit.cfg['fid_init_viewer'] = True
                if not qkit.module_available['pandas']:
                    raise ImportError("pandas not found. Pandas is needed for the fid viewer.\nInstall pandas or set qkit.cfg['fid_init_viewer']=False")
                self._initiate_basic_df()
            else:
                qkit.cfg['fid_init_viewer'] = False
    
    def _get_setting_from_set_file(self, filename, instrument, value):
        try:
            with open(filename, 'r') as f:
                i_name = False
                return_dict = {}
                for line in f:
                    if line.startswith("Instrument: "):
                        if line[12:].strip() in instrument:
                            i_name = line[12:].strip()
                        else:
                            i_name = False
                    elif i_name:
                        p = line.strip().split(":")
                        if p[0] in value:
                            for i in range(len(instrument)):
                                if p[0] == value[i] and i_name == instrument[i]:
                                    try:
                                        return_dict.update({i_name+":"+p[0] : float(p[1][1:])})
                                    except(ValueError,TypeError):
                                        return_dict.update({i_name + ":" + p[0]: p[1][1:]})
                return return_dict
        except(IOError):
            return {}

    def set_rating(self, uid, rating):
        """
        If you want to rate your measurements, so that you can filter for good ones. You can add a rating into the analysis
        folder of the h5 file
        :param uid: uid of the measurement file you wanna rate
        :type str
        :param rating: a simple value to rate your measurement. Default is 10, Rating <=0 is masked by default.
        :type int, float
        """
        return self._set_hdf_attribute(uid, "rating", rating)

    if qkit.module_available['pandas']:
    
        def _initiate_basic_df(self):
            """
            Creates a pandas data frame from your measurement
            data and allows to extract import values from h5-files
            """
        
            if len(self.h5_info_db) == 0:  # necessary if a data directory is chosen without any h5 file
                self.df = pd.DataFrame(columns=['datetime', 'name', 'run', 'user'])
            else:
                self.df = pd.DataFrame(self.h5_info_db).T
        
            if qkit.cfg.get('fid_scan_hdf', False):
                # self.df = self.df[['datetime', 'name', 'run', 'user', 'comment', 'fit_time', 'fit_freq', 'rating']]
                for key in ['rating', 'fit_time', 'fit_freq']:
                    if key in self.df.keys():
                        self.df[key] = pd.to_numeric(self.df[key], errors='coerce')
            else:
                self.df = self.df[['datetime', 'name', 'run', 'user']]
            self.df['datetime'] = pd.to_datetime(self.df['datetime'], errors='coerce')
            self.df.fillna("", inplace=True)  # Replace NAs with empty string to be able to detect changes
        
        def _get_settings_column(self, device, setting, uid=None, update_hdf=False):
            dfsetting = pd.DataFrame()
            if not isinstance(device, (tuple, list)):
                device = [device]
            if not isinstance(setting, (tuple, list)):
                setting = [setting]
            if len(device) != len(setting):
                raise ValueError("Please specify 'device' and 'setting' as equally long lists, where teir individual entries correspond to each other.")
            if uid is None:
                uid = self.df.index
            for i in uid:
                values = self._get_setting_from_set_file(self.h5_db[i].replace('.h5', '.set'), device, setting)
                if update_hdf:
                    for p, v in values.items():
                        self._set_hdf_attribute(i, p, v)
                dfsetting = pd.concat([dfsetting,
                                       pd.DataFrame(values, index=[i])
                                       ], sort=False)
            return dfsetting
    
        def add_settings_column(self, device, setting, measurement_id=None):
            """
            Reads out a specific setting from your chosen device. If you provide a uid,
            then only these files will be considered
            :param device: your device name
            :type str
            :param setting: setting of your device
            :type str
            :param uid: measurement_id (list). If None (default), all are used
            :type str
            """
            settings_column = self._get_settings_column(device, setting, measurement_id, update_hdf=qkit.cfg.get('fid_scan_hdf', False))
            for key in settings_column.keys():
                if key in self.df.keys():
                    self.df.update(settings_column.loc[:, key])
                else:
                    self.df = pd.concat([self.df, settings_column.loc[:, key]], axis=1)
        
        def open_in_filemanager(self):
            ids = self._selected_df.index
            if len(ids) > 10:
                logging.error("You are trying to open more than 10 files, are you sure?")
                return
            from platform import system
            _os = system()
            if _os == "Windows":
                from subprocess import Popen
                for i in ids:
                    Popen(r'explorer /select,"{}"'.format(self.h5_db[i]))
            else:
                logging.error("File Manager currently only supported for Windows")
                return
    
        def remove_column(self, column):
            """
            If your data frame is getting too wide, you can remove single columns
            :param column: column name of your data frame
            :type str
            """
            self.df = self.df.drop([column], axis=1)
    
        def _on_row_selected(self, row):
            self._selected_df = self.grid.get_selected_df()
    
        def _on_openSelected_clicked(self, b):
            uuid = self._selected_df.index
            logging.info("Open qviekit with uuid:%s" % uuid)
            self.view(uuid)
    
        def _grid_observer(self, change):
            logging.debug('grid observer called, batch_update: ' + str(self._batch_update))
            c = self.grid
            keys = [i for i in list(c.df.keys()) if i != "qgrid_unfiltered_index"]
            if self._batch_update:
                self._batch_update = False
                changed_df = c.df
            else:
                changed_df = c.get_changed_df()
            try:
                uuids = [i for i in list(changed_df.index) if i in self.df.index]
                indices = np.where(self.df.loc[uuids, keys] != changed_df.loc[uuids, keys])
                logging.debug("I found {} changes".format(len(indices[0])))
                for i in range(len(indices[0])):
                    index = uuids[indices[0][i]]
                    key = keys[indices[1][i]]
                    new_value = changed_df.loc[index, key]
                    if self.df.loc[index, key] != new_value:
                        self._set_hdf_attribute(index, key, new_value)
                        logging.debug("{}[{}] '{}'-> '{}'".format(index, key, self.df.loc[index, key], new_value))
                        self.df.loc[index, key] = new_value
            except(ValueError, KeyError):
                logging.info("Updating the measurment database failed.")
                self.debug = [self.df.copy(), changed_df, keys]
    
        def _batch_change_attribute(self, b):
            tmp = self.grid.get_changed_df()
            tmp.loc[self._selected_df.index, b.key_dd.value] = b.value_tf.value
            self._batch_update = True
            self.grid.df = tmp
    
        def show(self, show_raw=False):
            """
            used to show the data base as a qgrid object or if not installed pandas data frame
            :return: data frame as qgrid object or pandas object
            """
            self.wait()
            if self.found_qgrid:
                if LooseVersion(qd.__version__) < LooseVersion("1.3.0") and LooseVersion(pd.__version__) >= LooseVersion("1.0"):
                    logging.warning("qgrid < v1.3 is incompatible with pandas > v1.0. Check for a new version of qgrid or downgrade pandas to v0.25.3")
                    self.found_qgrid = False
            if self.found_qgrid:
                from IPython.display import display
                import ipywidgets as widgets
                rows = [d for d in self.column_sorting if d in list(self.df.keys()) and d not in self.columns_ignore]
                rows += [d for d in list(self.df.keys()) if d not in self.column_sorting and d not in self.columns_ignore]
            
                _openSelected = widgets.Button(description='open selected', disabled=False,
                                               button_style='', tooltip='open selected')
                _openSelected.on_click(self._on_openSelected_clicked)
            
                _batch_modifier = widgets.Button(description='for all selected rows',
                                                 tooltip='Change the selected attribute to the specified value for all rows selected.')
                _batch_modifier.key_dd = widgets.Dropdown(options=rows, description="Set")
                _batch_modifier.value_tf = widgets.Text(value='', placeholder='Value', description='to:')
                _batch_modifier.on_click(self._batch_change_attribute)
            
                display(widgets.HBox([_openSelected, _batch_modifier.key_dd, _batch_modifier.value_tf, _batch_modifier]))
            
                if show_raw or "rating" not in self.df.keys():
                    df = self.df.copy()
                else:
                    df = self.df.copy()[pd.to_numeric(self.df['rating']) > 0]
                self.grid = qd.show_grid(df[rows], show_toolbar=False, grid_options={'enableColumnReorder': True})
                self.grid.observe(self._on_row_selected, names=['_selected_rows'])
                self.grid.observe(self._grid_observer, names=[
                    '_selected_rows'])  # Quick fix to also work with newer versions of qgrid. Should be changed to the .on() event mechanism at some point, but this requires newer qgrid version.
                return self.grid
            else:
                if show_raw or "rating" not in self.df.keys():
                    return self.df
                else:
                    return self.df[self.df['rating'] > 0]
    
        def get_filtered_uuids(self):
            """
            Thsi function gives you a list of UUIDs in your current view.
            If you have selected more than one line, only the selection is returned.
            Otherwise, all measurements in the current view are given.
            This is especially handy if you did some filtering beforehand
            
            This of course only works if qgrid is installed.
            
            :return: list of UUIDs
            """
            if self.found_qgrid:
                if len(self._selected_df) > 1:
                    return list(self._selected_df.index)
                else:
                    return list(self.grid.get_changed_df().index)
            else:
                logging.warning("Module qgrid is not installed. Filtering the database is only supported with qgrid.")
                return False
        
        def search(self, column, expression=None, value=None, bounds=None):
            """
            allows you to search a specific column for either a string, a value, or values within bounds.
            You have to pass exactly one variable. If you are only looking for the uids write ".index" behind it.
            :param column: name of the column you want to search
            :type str
            :param expression: if you wanna search for a string use this
            :type str
            :param value: if you wanna search for a value use this
            :type int or float
            :param bounds: list of lower and upper bound
            :type int
            :return: pandas data frame where the values you are searching for are included
            """
            if expression is not None and value is None and bounds is None:
                return self.df[self.df[column].str.contains(expression, na=False)]
            if value is not None and expression is None and bounds is None:
                print([self.df[column] == value])
                return self.df[self.df[column] == value]
            if bounds is not None and value is None and expression is None:
                return self.df[((self.df[column] > bounds[0]) & (self.df[column] < bounds[1]))]
            else:
                print("You have to pass exactly one variable")
    
        def add_column(self, colname, value=" "):
            """
            Add a new (empty) column to your dataframe. This will be displayed in the view after qkit.fid.show() and you can manipulate the individual values there.
            :param colname: Name of your column. A ValueError is raised if this already exists.
            :param value: Default value for all entries. Should not be empty string '' but space ' ' as changes can not be detected otherwise.
            """
            if colname in self.df:
                raise ValueError("Column {} is already in your dataset. Can not be added twice".format(colname))
            else:
                self.df.loc[:, colname] = pd.Series(value, index=self.df.index)
    else:
        # If pandas is not installed, raise an error if the public functions are called
        
        def void_func(self):
            raise ImportError("This function requires pandas to be installed.")

        add_settings_column = open_in_filemanager = remove_column = show = get_filtered_uuids = search = add_column = void_func

    def enlarge_notebook(self,width=100):
        from IPython.core.display import display, HTML
        display(HTML("<style>.container { width:%i%% !important; }</style>"%width))


# edw_server.py by JB@KIT 03/2015
# TCP server using EdwardsBridgeReadout USB interface

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

try:
    import SocketServer
except ImportError:
    import socketserver as SocketServer

import os,sys
sys.path.append('../')
from EdwardsBridgeReadout import Edwards_p_gauge   #import readout class

p = Edwards_p_gauge()

class TCPHandler_Edwards(SocketServer.BaseRequestHandler):
    """
    The RequestHandler class for our server.

    It is instantiated once per connection to the server, and must
    override the handle() method to implement communication to the
    client.
    """

    def handle(self):
        try:
            # self.request is the TCP socket connected to the client
            self.data = self.request.recv(1024).strip()
            if str(self.data) == 'get_p_cond':
                print 'condenser line pressure request from', str(format(self.client_address[0]))
                self.request.sendall(str(p.get_condenser_pressure()))
            elif str(self.data) == 'get_p_still':
                print 'still line pressure request from', str(format(self.client_address[0]))
                self.request.sendall(str(p.get_still_pressure()))
            else:
                print 'request string not recognized'
                self.request.sendall(0)
        except Exception as m:
            print 'Error in TCP server handler:', m

if __name__ == "__main__":
    HOST, PORT = 'ip-address', 9955

    # Create the server, binding to localhost on port 9955
    server = SocketServer.TCPServer((HOST, PORT), TCPHandler_Edwards)

    # Activate the server; this will keep running until you
    # interrupt the program with Ctrl-C
    server.serve_forever()




try:
    import numpy as np
    import time
    import ipywidgets as widgets
    from IPython.display import display
    
    class Stopwatch(object):
        def __init__(self):
            self.sw = widgets.Label(value=time.strftime('%H:%M:%S', time.gmtime(0)),
                                    description='elapsed time',
                                    #layout=widgets.Layout(**{}),
                                    #style={},
                                    disabled=True)
            display(self.sw)
        
        def update(self, value):
            if type(value) is str:
                self.sw.value = value
            else:
                self.sw.value = time.strftime('%H:%M:%S', time.gmtime(value))
            
                    
except ImportError as e:
    print('ipywidgets could not be imported')


'''
generate_waveform.py
M. Jerger, S. Probst, A. Schneider (04/2015), J. Braumueller (04/2016)
'''

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import numpy as np
#import os.path
import time
import logging
import numpy
import sys
import qkit
if qkit.module_available("scipy"):
    import scipy.special

# creates dummy objects to have everything well-defined.
#global sample
#sample = type('Sample', (object,),{  'exc_T' : 1e-6 , 'tpi' : 2e-9 , 'tpi2' : 1e-9, 'clock' : 1e9   })

dtype = np.float16 #you can change this via gwf.dtype to anything you want

def compensate(wfm, gamma, sample):
    '''
    Function that translates a given (analog) waveform wfm into a waveform wfc that needs to be programmed to the AWG
    to effectively obtain the waveform wfm after the bias T.
    This compensation is required due to the finite time constant of the bias Ts capacitor. The time constant in seconds 
    is passed to the function as gamma.
    Credits to A. Schneider
    
    Inputs:
        - wfm: original waveform to be compensated
        - gamma: bias T time constant in seconds
        - sample: sample object form which the function reads the AWG clock
    Outputs:
        - wfc: corrected waveform to be loaded to the AWG
    '''
    wfc = np.zeros_like(wfm)
    dif = np.diff(wfm)
    for i in range(1,len(wfm)):
        wfc[i]=wfc[i-1]+wfm[i-1]/(sample.clock*gamma)+dif[i-1]
    return wfc


def erf(pulse, attack, decay, sample, length=None, position = None, low=0, high=1, clock = None):
    '''
        create an erf-shaped envelope function
        erf(\pm 2) is almost 0/1, erf(\pm 1) is ~15/85%
        
        Input:
            tstart, tstop - erf(-2) times
            attack, decay - attack/decay times
    '''
    if not qkit.module_available("scipy"):
        raise ImportError('scipy not available. scipy is needed for erf.')
    if(clock == None): clock = sample.clock
    if(length == None): length = sample.exc_T
    if position == None:   #automatically correct overlap only when position argument not explicitly given
        position = length
        if hasattr(sample, 'overlap'):   #if overlap exists in sample object
            position -= sample.overlap
        else:
            logging.warning('overlap attribute not found in sample object')
    if(pulse>position):
        logging.error(__name__ + ' : pulse does not fit into waveform')
        
    sample_start = int(clock*(position-pulse))
    sample_end = int(clock*position)
    sample_length = int(np.round(length*clock))
    wfm = low * np.ones(sample_length,dtype=dtype)
    
    if attack != 0:
        if attack < 2./clock:
            logging.warning(__name__ + ' : attack too small compared to AWG sample frequency, setting to %.4g s'%(2./clock))
            attack = 2./clock
        nAttack = int(clock*attack)
        sAttack = 0.5*(1+scipy.special.erf(np.linspace(-2, 2, nAttack)))
        wfm[sample_start:sample_start+nAttack] += sAttack * (high-low)
    else:
        nAttack = 0
    if decay != 0:
        if decay < 2./clock:
            logging.warning(__name__ + ' : decay too small compared to AWG sample frequency, setting to %.4g s'%(2./clock))
            decay = 2./clock
        nDecay = int(clock*decay)
        sDecay = 0.5*(1+scipy.special.erf(np.linspace(2, -2, nDecay)))
        wfm[sample_end-nDecay:sample_end] += sDecay * (high-low)
    else:
        nDecay = 0
    wfm[sample_start+nAttack:sample_end-nDecay] = high
    return wfm
    
def exp(pulse, decay, sample, position = None, low=0, high=1, clock = None):
    '''
    create and exponential decaying waveform
    '''
    if(clock == None): clock = sample.clock
    if position == None:   #automatically correct overlap only when position argument not explicitly given
        position = length
        if hasattr(sample, 'overlap'):   #if overlap exists in sample object
            position -= sample.overlap
        else:
            logging.warning('overlap attribute not found in sample object')
    sample_length = int(np.ceil(sample.exc_T*clock))
    wfm = low * np.ones(sample_length,dtype=dtype)
    sample_start = int(clock*(position-pulse))
    sample_end = int(clock*position)
    
    wfm[sample_start:sample_end] += np.exp(-np.arange(sample_end-sample_start)/(decay*clock)) * (high-low)
    return wfm
    
def triangle(attack, decay, sample, length = None, position = None, low=0, high=1, clock = None):
    '''
        create a pulse with triangular shape
        
        Input:
            attack, decay - attack and decay times in sec
            length - length of the comlete resulting waveform in sec
            position - position of the end of the pulse
    '''
    if(clock == None): clock = sample.clock
    if(length == None): length = sample.exc_T
    if position == None:   #automatically correct overlap only when position argument not explicitly given
        position = length
        if hasattr(sample, 'overlap'):   #if overlap exists in sample object
            position -= sample.overlap
        else:
            logging.warning('overlap attribute not found in sample object')
    sample_start = int(clock*(position-attack-decay))
    sample_end = int(clock*position)
    sample_length = int(np.ceil(length*clock))
    sample_attack = int(np.ceil(attack*clock)) 
    sample_decay = int(np.ceil(decay*clock))
    wfm = low * np.ones(sample_length,dtype=dtype)
    wfm[sample_start:sample_start+sample_attack] = np.linspace(low, high, sample_attack)
    wfm[sample_start+sample_attack:sample_end-sample_decay] = high
    wfm[sample_end-sample_decay:sample_end] = np.linspace(high, low, sample_decay)
    return wfm

def square(pulse, sample, length = None,position = None, low = 0, high = 1, clock = None, adddelay=0.,freq=None):
    '''
        generate waveform corresponding to a dc pulse

        Input:
            pulse - pulse duration in seconds
            length - length of the generated waveform
            position - time instant of the end of the pulse
            low - pulse 'off' sample value
            high - pulse 'on' sample value
            clock - sample rate of the DAC
        Output:
            float array of samples
    '''
    if(clock == None): clock= sample.clock
    if(length == None): length= sample.exc_T
    if position == None:   #automatically correct overlap only when position argument not explicitly given
        position = length
        if hasattr(sample, 'overlap'):   #if overlap exists in sample object
            position -= sample.overlap
        else:
            logging.warning('overlap attribute not found in sample object')
    if(pulse>position): logging.error(__name__ + ' : pulse does not fit into waveform')
    sample_start = int(clock*(position-pulse-adddelay))
    sample_end = int(clock*(position-adddelay))
    sample_length = int(np.round(length*clock)/4)*4 #Ensures that the number of samples is divisible by 4 @andre20150615
    #sample_length = int(np.ceil(length*clock)) #old definition
    wfm = low*np.ones(sample_length,dtype=dtype)
    if(sample_start < sample_end): wfm[int(sample_start)] = high + (low-high)*(sample_start-int(sample_start))
    if freq==None: wfm[int(np.ceil(sample_start)):int(sample_end)] = high
    else:
        for i in range(int(sample_end)-int(np.ceil(sample_start))):
            wfm[i+int(np.ceil(sample_start))] = high*np.sin(2*np.pi*freq/clock*(i))
    if(np.ceil(sample_end) != np.floor(sample_end)): wfm[int(sample_end)] = low + (high-low)*(sample_end-int(sample_end))
    return wfm
    
    
def gauss(pulse, sample, length = None,position = None, low = 0, high = 1, clock = None):
    '''
        generate waveform corresponding to a dc gauss pulse

        Input:
            pulse - pulse duration in seconds
            length - length of the generated waveform
            position - time instant of the end of the pulse
            low - pulse 'off' sample value
            high - pulse 'on' sample value
            clock - sample rate of the DAC
        Output:
            float array of samples
    '''
    if(clock == None): clock= sample.clock
    if(length == None): 
        length= sample.exc_T
        sample_length = int(np.round(length*clock)/4)*4
    else:
        sample_length = int(length*clock)
    if position == None:   #automatically correct overlap only when position argument not explicitly given
        position = length
        if hasattr(sample, 'overlap'):   #if overlap exists in sample object
            position -= sample.overlap
        else:
            logging.warning('overlap attribute not found in sample object')
    if(pulse>position): logging.error(__name__ + ' : pulse does not fit into waveform')
    sample_start = int(clock*(position-pulse))
    sample_end = int(clock*position)
    
    wfm = low*np.ones(sample_length,dtype=dtype)
    if(sample_start < sample_end): wfm[int(sample_start)] = 0.#high + (low-high)*(sample_start-int(sample_start))
    #wfm[int(np.ceil(sample_start)):int(sample_end)] = high
    pulsesamples = int(int(sample_end)-int(sample_start))
    for i in range(pulsesamples):
        wfm[int(np.ceil(sample_start)+i)] = high*np.exp(-(i-pulsesamples/2.)**2/(2.*(pulsesamples/5.)**2))
    if(np.ceil(sample_end) != np.floor(sample_end)): wfm[int(sample_end)] = low + (high-low)*(sample_end-int(sample_end))
    return wfm

def arb_function(function, pulse, length = None,position = None, clock = None):
    '''
        generate arbitrary waveform pulse

        Input:
            function - function called
            pulse - duration of the signal
            position - time instant of the end of the signal
            length - duration of the waveform
            clock - sample rate of the DAC
        Output:
            float array of samples
    '''
    if(clock == None): clock= sample.clock
    if(length == None): length= sample.exc_T
    if position == None:   #automatically correct overlap only when position argument not explicitly given
        position = length
        if hasattr(sample, 'overlap'):   #if overlap exists in sample object
            position -= sample.overlap
        else:
            logging.warning('overlap attribute not found in sample object')
    if(pulse>position): logging.error(__name__ + ' : pulse does not fit into waveform')
    sample_start = clock*(position-pulse)
    sample_end = clock*position
    sample_length = int(np.ceil(length*clock))

    wfm = np.zeros(sample_length,dtype=dtype)
    times = 1./clock*np.arange(0, sample_end-sample_start+1)
    wfm[sample_start:sample_end] = function(times)
    return wfm

    
def t1(delay, sample, length = None, low = 0, high = 1, clock = None, DRAG_amplitude=None):
    '''
        generate waveform with one pi pulse and delay after

        Input:
            delay - time delay after pi pulse
            sample object

        Output:
            float array of samples
    '''
    if(clock == None): clock = sample.clock
    if(length == None): length = sample.exc_T

    if hasattr(sample, 'overlap'):   #if overlap exists in sample object
        delay += sample.overlap
    else:
        logging.warning('overlap attribute not found in sample object')

    if(delay+sample.tpi > length): logging.error(__name__ + ' : pulse does not fit into waveform')
    
    if DRAG_amplitude == None:
        wfm = square(sample.tpi, sample, length, length-delay, clock = clock)
    else:
        wfm = drag(sample.tpi, sample, DRAG_amplitude, length, length-delay, clock = clock)
    wfm = wfm * (high-low) + low
    return wfm
    
    
def ramsey(delay, sample, pi2_pulse = None, length = None,position = None, low = 0, high = 1, clock = None, DRAG_amplitude=None):
    '''
        generate waveform with two pi/2 pulses and delay in-between

        Input:
            delay - time delay between the pi/2 pulses
            pi2_pulse - length of a pi/2 pulse
            (see awg_pulse for rest)
        Output:
            float array of samples
    '''
    if(clock == None): clock = sample.clock
    if(length == None): length = sample.exc_T
    if position == None:   #automatically correct overlap only when position argument not explicitly given
        position = length
        if hasattr(sample, 'overlap'):   #if overlap exists in sample object
            position -= sample.overlap
        else:
            logging.warning('overlap attribute not found in sample object')
    if DRAG_amplitude == None:
        if(pi2_pulse == None): pi2_pulse = sample.tpi2
        if(delay+2*pi2_pulse>position): logging.error(__name__ + ' : ramsey pulses do not fit into waveform')
        wfm = square(pi2_pulse, sample, length, position, clock = clock)
        wfm += square(pi2_pulse, sample,  length, position-delay-pi2_pulse, clock = clock)
    
    else:
        if(pi2_pulse == None): pi2_pulse = sample.tpi2
        if(delay+2*pi2_pulse>position): logging.error(__name__ + ' : ramsey pulses do not fit into waveform')
        wfm = drag(pi2_pulse, sample, DRAG_amplitude, length, position, clock = clock)
        wfm += drag(pi2_pulse, sample, DRAG_amplitude,  length, position-delay-pi2_pulse, clock = clock)
    
    wfm = wfm * (high-low) + low
    return wfm


def spinecho(delay, sample, pi2_pulse = None, pi_pulse = None, length = None,position = None, low = 0, high = 1, clock = None, readoutpulse=True, adddelay=0., freq=None, n = 1,DRAG_amplitude=None, phase = 0.):
    '''
        generate waveform with two pi/2 pulses at the ends and a number n of echo (pi) pulses in between
        pi2 - delay/(n/2) - pi - delay/n - pi - ... - pi - delay/(n/2) - [pi2, if readoutpulse]
        
        Phase shift included to perform CPMG-Measurements,
        DRAG included
        Sequence for n>1 fixed: between pi2 and pi is delay/(n/2) @TW20160907
        
        pulse - pulse duration in seconds
        length - length of the generated waveform
        position - time instant of the end of the pulse
        low - pulse 'off' sample value
        high - pulse 'on' sample value
        clock - sample rate of the DAC
        phase - phase shift between pi2 and pi pulses in rad
        DRAG_amplitude - if not None, DRAG-Pulses are used
        
        waveforms are contructed from right to left
    '''
    
    if(clock == None): clock= sample.clock
    if(length == None): length= sample.exc_T
    if position == None:   #automatically correct overlap only when position argument not explicitly given
        position = length
        if hasattr(sample, 'overlap'):   #if overlap exists in sample object
            position -= sample.overlap
        else:
            logging.warning('overlap attribute not found in sample object')
    if(pi2_pulse == None): pi2_pulse = sample.tpi2
    if(pi_pulse == None): pi_pulse = sample.tpi
  
    if round(adddelay+delay+2*pi2_pulse+n*pi_pulse, 10) > round(position, 10): # round bc of floating points arethmetic
        logging.error(__name__ + ' : sequence does not fit into waveform. delay is the sum of the waiting times in between the pi pulses')
    
    if DRAG_amplitude == None:
        if readoutpulse:   #last pi/2 pulse
            wfm = square(pi2_pulse, sample, length, position, clock = clock, freq=freq)*np.exp(0j)   #add pi/2 pulse
        else:
            wfm = square(pi2_pulse, sample, length, position, low, low, clock,freq=freq)*np.exp(0j)   #create space (low) of the length of a pi/2 pulse
        for ni in range(n):   #add pi pulses
            wfm += square(pi_pulse, sample, length, position - pi2_pulse - ni*pi_pulse - float(delay)/(2*n)-delay/n*ni - adddelay, clock = clock, freq=freq)*np.exp(phase*1j)
        wfm += square(pi2_pulse, sample, length, position - pi2_pulse - n*pi_pulse - delay - adddelay, clock = clock, freq=freq)*np.exp(0j)   #pi/2 pulse
        wfm = wfm * (high-low) + complex(low,low)   #adjust offset
        if phase == 0: wfm = wfm.real # to avoid conversion error messages
        
    else:
        if readoutpulse:   #last pi/2 pulse
            wfm = drag(pi2_pulse, sample, DRAG_amplitude, length, position, clock = clock) *np.exp(0j)   #add pi/2 pulse
        else:
            wfm = square(pi2_pulse, sample, length, position, low, low, clock,freq=freq)*np.exp(0j)   #create space (low) of the length of a pi/2 pulse
        for ni in range(n):   #add pi pulses
            wfm += drag(pi_pulse, sample, DRAG_amplitude, length, position - pi2_pulse - ni*pi_pulse - float(delay)/(2*n)-delay/n*ni - adddelay, clock = clock)*np.exp(phase*1j)
        wfm += drag(pi2_pulse, sample, DRAG_amplitude, length, position - pi2_pulse - n*pi_pulse - delay - adddelay, clock = clock)*np.exp(0j)   #pi/2 pulse
        wfm = wfm * (high-low) + complex(low,low)   #adjust offset
    
    return wfm
    
def udd(delay, sample, pi2_pulse = None, pi_pulse = None, length = None,position = None, low = 0, high = 1, clock = None, readoutpulse=True,adddelay=0., freq=None, n = 1, DRAG_amplitude=None, phase = np.pi/2): 
    '''
        generate waveform with two pi/2 pulses at the ends and a number n of (pi) pulses in between
        where the position of the j-th pulse is defined by sin^2[(pi*j)/(2N+2)]  @TW20160908
       
        pulse - pulse duration in seconds
        length - length of the generated waveform
        position - time instant of the end of the pulse
        low - pulse 'off' sample value
        high - pulse 'on' sample value
        clock - sample rate of the DAC
        phase - phase shift between pi2 and pi pulses
        DRAG_amplitude - if not None, DRAG-Pulses are used
    '''
    
    if(clock == None): clock= sample.clock
    if(length == None): length= sample.exc_T
    if position == None:   #automatically correct overlap only when position argument not explicitly given
        position = length
        if hasattr(sample, 'overlap'):   #if overlap exists in sample object
            position -= sample.overlap
        else:
            logging.warning('overlap attribute not found in sample object')
    if(pi2_pulse == None): pi2_pulse = sample.tpi2
    if(pi_pulse == None): pi_pulse = sample.tpi   
    
    if round(adddelay+delay+2*pi2_pulse+n*pi_pulse, 10) > round(position, 10):
        logging.error(__name__ + ' : sequence does not fit into waveform. delay is the sum of the waiting times in between the pi pulses')
    if DRAG_amplitude == None:
        if readoutpulse:   #last pi/2 pulse
            wfm = square(pi2_pulse, sample, length, position, clock = clock, freq=freq)*np.exp(0j)   #add pi/2 pulse
        else:
            wfm = square(pi2_pulse, sample, length, position, low, low, clock,freq=freq)*np.exp(0j)   #create space (low) of the length of a pi/2 pulse
        for ni in range(n):   #add pi pulses
            wfm += square(pi_pulse, sample, length, position - (delay+n*pi_pulse)*(np.sin((np.pi*(ni+1))/(2*n+2)))**2 - adddelay, clock = clock, freq=freq)*np.exp(phase*1j) # no pi2_pulse subtracted because equation yields position of center
        wfm += square(pi2_pulse, sample, length, position - pi2_pulse - n*pi_pulse - delay - adddelay, clock = clock, freq=freq)*np.exp(0j)   #pi/2 pulse
        wfm = wfm * (high-low) + complex(low,low)   #adjust offset
        if phase == 0: wfm = wfm.real # to avoid conversion error messages
    
    else:
        if readoutpulse:   #last pi/2 pulse
            wfm = drag(pi2_pulse, sample, DRAG_amplitude, length, position, clock = clock)*np.exp(0j)   #add pi/2 pulse
        else:
            wfm = square(pi2_pulse, sample, length, position, low, low, clock,freq=freq)*np.exp(0j)   #create space (low) of the length of a pi/2 pulse
        for ni in range(n):   #add pi pulses
            wfm += drag(pi_pulse, sample, DRAG_amplitude, length, position - (delay+n*pi_pulse)*(np.sin((np.pi*(ni+1))/(2*n+2)))**2 - adddelay, clock = clock)*np.exp(phase*1j)
        wfm += drag(pi2_pulse, sample, DRAG_amplitude, length, position - pi2_pulse - n*pi_pulse - delay - adddelay, clock = clock)*np.exp(0j)   #pi/2 pulse
        wfm = wfm * (high-low) + complex(low,low)   #adjust offset
    return wfm


def drag(pulse, sample, amplitude, length = None, position=None, clock = None):
    '''
        if pulses are short, DRAG helps to reduce the gate error.
        Pulseshape on I is gussian and on Q is the derivative of I times an experimentally determined amplitude
    
        pulse - pulse duration in seconds
        amplitude - experimentally determined amplitude
        length - length of the generated waveform
        position - time instant of the end of the pulse
        clock - sample rate of the DAC        
    '''

    if(clock == None): clock = sample.clock
    if(length == None): length = sample.exc_T
    if position == None:   #automatically correct overlap only when position argument not explicitly given
        position = length
        if hasattr(sample, 'overlap'):   #if overlap exists in sample object
            position -= sample.overlap
        else:
            logging.warning('overlap attribute not found in sample object')
    wfm = gauss(pulse, sample, length=np.ceil(length*1e9)/1e9, position=position) + 1j * np.concatenate([np.diff(gauss(pulse, sample,length=np.ceil(length*1e9)/1e9, position=position)*amplitude),[0]]) # actual pulse
    wfm[int((position-pulse)*clock-1):int((position-pulse)*clock+1)]=wfm.real[int((position-pulse)*clock-1):int((position-pulse)*clock+1)] # for smooth derivative
    wfm[int(position*clock-1):int(position*clock+1)]= wfm.real[int(position*clock-1):int(position*clock+1)] 
    
    return wfm



# Base class for QKIT measurements to simplify and unify
# S1@KIT 2020

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA


import logging
import threading

import numpy as np

import qkit
import qkit.measure.write_additional_files as waf
from qkit.gui.notebook.Progress_Bar import Progress_Bar
from qkit.gui.plot import plot as qviewkit
from qkit.measure.measurement_class import Measurement
from qkit.storage import store as hdf


##################################################################

class MeasureBase(object):
    """
    Baseclass for measurements.
    """

    def __init__(self, sample=None):
        self.set_sample(sample)
        
        self.comment = ''
        self.measurement_name = None
        
        self._x_parameter = None
        self._y_parameter = None
        
        self.progress_bar = True
        self._pb = Progress_Bar(0, dummy=True)
        
        self._file_name = ""
        
        self.log_functions = []
        
        self.open_qviewkit = True
        self.qviewkit_singleInstance = False
        self._qvk_process = False
        
        self._measurement_object = Measurement()
        self._measurement_object.measurement_type = 'defaultMeasurement'
        self.web_visible = True

    def __getattr__(self, item):
        if item == "x_vec":
            return self._x_parameter.values
        elif item == "y_vec":
            return self._y_parameter.values
        else:
            raise AttributeError("Measurement object has no attribute '{}'".format(item))
    
    def set_sample(self,sample):
        if sample is None:
            self._sample = None
        elif not isinstance(sample,qkit.measure.samples_class.Sample):
            raise ValueError(__name__ + ": The given sample '{!s}' is not a qkit sample object.".format(sample))
        self._sample = sample
    
    def add_log_function(self, func, name, unit="", log_dtype=float):
        """
        A function (object) can be passed to the measurement loop which is excecuted before every x iteration
        but after executing the x_object setter in 2D measurements and before every line (but after setting
        the x value) in 3D measurements.
        The return value of the function of type float or similar is stored in a value vector in the h5 file.
        
        To add multiple log functions, execute this function multiple times.

        func: function object
        name: name of logging parameter appearing in h5 file
        unit: unit of logging parameter, default: ''
        log_dtype: h5 data type, default: float
        """
        
        if not callable(func):
            raise ValueError('{:s}: Cannot set {!s} as log-function: callable object needed'.format(__name__, func))
        if type(name) is not str:
            raise ValueError('{:s}: Cannot set {!s} as log-name: string needed'.format(__name__, name))
        if type(unit) is not str:
            raise ValueError('{:s}: Cannot set {!s} as log-unit: string needed'.format(__name__, unit))
        
        self.log_functions.append([func, name, unit, log_dtype])
    
    def remove_log_function(self, index=None):
        """
        Remove the index-th log function.
        If no index is given, all log functions are removed.
        """
        if index is None:
            self.log_functions = []
        else:
            self.log_functions.pop(index)
    
    class Coordinate:
        def __init__(self, name, unit="", values=None, set_function=None, wait_time=0):
            if type(name) is not str:
                raise TypeError('{:s}: Cannot set {!s} as name for coordinate: string needed'.format(__name__, self.name))
            self.name = name
            self.unit = unit
            self.values = values
            if callable(set_function):
                self.set_function = set_function
            else:
                self.set_function = lambda x: True
            self.wait_time = wait_time if wait_time is not None else 0
            self.hdf_dataset = None
        
        def validate_parameters(self):
            if type(self.name) is not str:
                raise TypeError('{:s}: Cannot set {!s} as name for coordinate: string needed'.format(__name__, self.name))
            if type(self.unit) is not str:
                raise TypeError('{:s}: Cannot set {!s} as unit for coordinate {!s}: string needed'.format(__name__, self.unit, self.name))
            if np.iterable(self.values):
                try:
                    self.values = np.array(self.values, dtype=float)
                    if len(self.values) == 0:
                        logging.warning("{:s}: Setting zero-length vector for coordinate {!s}. Please check if this is intended.".format(__name__, self.name))
                except Exception as e:
                    raise type(e)(
                            '{:s}: Cannot set {!s} as vector for coordinate {!s}: Array conversion failed:{!s}'.format(__name__, self.values, self.name, e))
            else:
                raise TypeError('{:s}: Cannot set {!s} as vector for coordinate {!s}: iterable object needed'.format(__name__, self.values, self.name))
            if not callable(self.set_function):
                raise TypeError('{:s}:  Cannot set {!s} as function for coordinate {!s}: callable object needed'.format(__name__, self.set_function, self.name))
            if type(self.wait_time) in [int, float]:
                self.wait_time = float(self.wait_time)
            else:
                raise TypeError('{:s}:  Cannot set {!s} as wait time for coordinate {!s}: float needed'.format(__name__, self.wait_time, self.name))
            return True

        def create_dataset(self, hdf_file):
            if self.hdf_dataset is None or self.hdf_dataset.hf != hdf_file.hf: # If dataset not yet created or belongs to old hdf file
                self.validate_parameters()
                self.hdf_dataset = hdf_file.add_coordinate(self.name, unit=self.unit)
                self.hdf_dataset.add(self.values)
            else:
                logging.info(__name__ + ": Dataset for coordinate '{}' was already created.".format(self.name))
            return self.hdf_dataset

    class Data:
        def __init__(self, name, coords, unit="", save_timestamp=False, average_over=None, **kwargs):
            if type(name) is not str:
                raise TypeError('{:s}: Cannot set {!s} as name for Data: string needed'.format(__name__, name))
            if type(coords) not in (list, tuple):
                raise TypeError('{:s}: Cannot set {!s} as coordinates for data {!s}: list or tuple of Coordinates needed'.format(__name__, coords, name))
            for c in coords:
                if not isinstance(c,MeasureBase.Coordinate):
                    raise TypeError(
                            '{:s}: Cannot set {!s} as coordinate for data {!s}: Please use the qkit.measure.MeasureBase.Coordinate type'.format(__name__, c,
                                                                                                                                                name))
            self.name = name
            self.coordinates = coords
            self.dim = len(coords)
            self.unit = unit
            self.kwargs = kwargs
            self.kwargs.update({"save_timestamp": save_timestamp, "average_over": average_over})
            self.hdf_dataset = None
    
        def validate_parameters(self):
            if type(self.name) is not str:
                raise TypeError('{:s}: Cannot set {!s} as name for Data: string needed'.format(__name__, self.name))
            if type(self.coordinates) not in (list, tuple):
                raise TypeError(
                        '{:s}: Cannot set {!s} as coordinates for data {!s}: list or tuple of Coordinates needed'.format(__name__, self.coordinates, self.name))
            for c in self.coordinates:
                if not isinstance(c,MeasureBase.Coordinate):
                    raise TypeError(
                            '{:s}: Cannot set {!s} as coordinate for data {!s}: Please use the qkit.measure.MeasureBase.Coordinate type'.format(__name__, c,
                                                                                                                                                self.name))
            self.dim = len(self.coordinates)
            if type(self.unit) is not str:
                raise TypeError('{:s}: Cannot set {!s} as unit for data {!s}: string needed'.format(__name__, self.unit, self.name))
            return True
    
        def create_dataset(self, hdf_file):
            if self.hdf_dataset is None or self.hdf_dataset.hf != hdf_file.hf: # If dataset not yet created or belongs to old hdf file
                self.validate_parameters()
                c = [co.create_dataset(hdf_file) for co in self.coordinates]
                if self.dim == 1:
                    self.hdf_dataset = hdf_file.add_value_vector(self.name, x=c[0], unit = self.unit, **self.kwargs)
                elif self.dim == 2:
                    self.hdf_dataset = hdf_file.add_value_matrix(self.name, x=c[0], y=c[1], unit = self.unit, **self.kwargs)
                elif self.dim == 3:
                    self.hdf_dataset = hdf_file.add_value_box(self.name, x=c[0], y=c[1], z=c[2], unit = self.unit, **self.kwargs)
            else:
                logging.info(__name__ + ": Dataset for coordinate '{}' was already created.".format(self.name))
            return self.hdf_dataset
    
    def set_x_parameters(self, vec, coordname, set_obj, unit, dt=None):
        """
        Sets x-parameters for 2D and 3D scan.
        In a 3D measurement, the x-parameters will be the "outer" sweep meaning for every x value all y values are swept and for each (x,y) value the bias is swept according to the set sweep parameters.

        Parameters
        ----------
        vec: array_likes
            An N-dimensional array that contains the sweep values.
        coordname: string
            The coordinate name to be created as data series in the .h5 file.
        set_obj: obj
            An callable object to execute with vec-values.
        unit: string
            The unit name to be used in data series in the .h5 file.
        dt: float, optional
            The sleep time between x-iterations.

        Returns
        -------
        None
        """
        # vec
        try:
            self._x_parameter = self.Coordinate(coordname, unit, np.array(vec, dtype=float), set_obj, dt)
            self._x_parameter.validate_parameters()
        except Exception as e:
            self._x_parameter = None
            raise e
    
    def set_y_parameters(self, vec, coordname, set_obj, unit, dt=None):
        """
        Sets y-parameters for 2D and 3D scan.
        In a 3D measurement, the y-parameters will be the "outer" sweep meaning for every y value all y values are swept and for each (y,y) value the bias is swept according to the set sweep parameters.

        Parameters
        ----------
        vec: array_likes
            An N-dimensional array that contains the sweep values.
        coordname: string
            The coordinate name to be created as data series in the .h5 file.
        set_obj: obj
            An callable object to execute with vec-values.
        unit: string
            The unit name to be used in data series in the .h5 file.
        dt: float, optional
            The sleep time between y-iterations.

        Returns
        -------
        None
        """
        # vec
        try:
            self._y_parameter = self.Coordinate(coordname, unit, np.array(vec, dtype=float), set_obj, dt)
            self._y_parameter.validate_parameters()
        except Exception as e:
            self._y_parameter = None
            raise e
    
    def _prepare_measurement_devices(self):
        """
        Perform all relevant device settings.
        This is a placeholder function and should be overwritten in the individual measurement scripts.
        """
        pass

    def _create_file_name(self, data):
        coordinates = set()
        self._dim = 0
        for d in data:
            for c in d.coordinates:
                coordinates.add(c.name.rstrip("_0123456789")) # removes indices like frequency_0 from the file name
            self._dim = max(self._dim, len(d.coordinates))  # if you have several 2D scans, the dimension is still 2D
        if not self.measurement_name:
            self.measurement_name = ", ".join(coordinates)
        self._file_name = '{}D_'.format(self._dim) + self.measurement_name.replace(' ', '').replace(',', '_')

    def _prepare_measurement_file(self, data, coords=()):
        """
        creates the output .h5-file with distinct dataset structures for each measurement type.
        at this point all measurement parameters are known and put in the output file
        All nacessary coordinates are alread included in the data instances, but you can supply a list of additional coords, which will be created.
        """
        for c in coords:
            if not isinstance(c, self.Coordinate):
                raise TypeError('{:s}:  {!s} is no valid coordinate object'.format(__name__, c))
        self._create_file_name(data)
        
        self._data_file = hdf.Data(name=self._file_name, mode='a')
        self._datasets = {}
        self._coordinates = {}
        for d in data:
            self._datasets[d.name] = d.create_dataset(self._data_file)
            for c in d.coordinates:
                self._coordinates[c.name] = c.create_dataset(self._data_file)  # it doesn't matter if you call create_dataset multiple times.
        for c in coords:
            self._coordinates[c.name] = c.create_dataset(self._data_file)  # the additional coordinates

        self._measurement_object.coordinates = list(self._coordinates.keys())
        self._measurement_object.data = list(self._datasets.keys())
        self._measurement_object.web_visible = self.web_visible
        
        self._measurement_object.instruments = qkit.instruments.get_instrument_names()
        self._measurement_object.sample = self._sample
        self._measurement_object.write_to_hdf(self._data_file)
        self._measurement_object.save()
        
        # write logfile and instrument settings
        self._settings = self._data_file.add_textlist('settings')
        settings = waf.get_instrument_settings(self._data_file.get_filepath())
        self._settings.append(settings)
        self._log = waf.open_log_file(self._data_file.get_filepath())

        if self._dim > 1:
            self._log_datasets = []
            for [func, name, unit, log_dtype] in self.log_functions:
                self._log_datasets.append([self._data_file.add_value_vector(name, x=data[0].coordinates[0].create_dataset(self._data_file), unit=unit,
                                                                            dtype=log_dtype), func])
        
        if self.comment:
            self._data_file.add_comment(self.comment)
    
    def _open_qviewkit(self, datasets=None):
        """
        Closes the old qvk process if needed and creates a new one.
        You can specify e.g. datasets=['amplitude','phase'] which will be openend by default and saved as an attribute to the h5 file
        as default, the self._datasets you created are opened.
        """
        if self.qviewkit_singleInstance and self.open_qviewkit and self._qvk_process:
            self._qvk_process.terminate()  # terminate an old qviewkit instance
        
        if 'default_ds' in self._data_file.hf.hf.attrs and datasets is None:
            datasets = list(self._data_file.hf.hf.attrs['default_ds'])
        if datasets is None:
            datasets = list(self._datasets.keys())
        self._data_file.hf.hf.attrs['default_ds'] = datasets
        
        if self.open_qviewkit:
            self._qvk_process = qviewkit.plot(self._data_file.get_filepath(), datasets=datasets)
    
    def _acquire_log_functions(self):
        for [ds, func] in self._log_datasets:
            ds.append([func()])
    
    def _end_measurement(self):
        """
        the data file is closed and filepath is printed
        """
        print(self._data_file.get_filepath())
        threading.Thread(target=qviewkit.save_plots, args=[self._data_file.get_filepath()]).start()
        self._data_file.close_file()
        waf.close_log_file(self._log)
        self.measurement_name = None
        qkit.flow.end()



# -*- coding: utf-8 -*-
"""
Created 2015

@author: hrotzing
"""


import logging
import numpy
import time
import qkit
from qkit.storage.hdf_constants import ds_types
from qkit.measure.json_handler import QkitJSONEncoder, QkitJSONDecoder

class hdf_dataset(object):
    """Dataset representation in qkit.        
    
    Wrapper class around the native h5py.crate_dataset().
    This is also a helper class to postpone the creation of the datasets.
    The main issue here is that until the first data is aquired, 
    some dimensions and items are unknown.
    To keep the userinterface simple, we choose to postpone the creation 
    of the datasets and derive all the unknown values from the real data.
    The working horse here is the 'append()' or the 'add()' function. Before, 
    just an empty dataset is created and the metadata are set.
    """
    
    def __init__(self, hdf_file, name='', 
                 ds_url=None, 
                 x=None, 
                 y=None, 
                 z=None, 
                 unit= "", 
                 comment="",
                 folder = 'data', 
                 save_timestamp = False, 
                 overwrite=False,
                 ds_type = ds_types['vector'],
                 **meta):
        """Init the dataset object with case sensitive arguments"""
        name = name.lower().replace(" ","_")
        self.hf = hdf_file
        self.x_object = x
        self.y_object = y
        self.z_object = z
        self.meta = meta
        self.dim = self.meta.pop('dim', None)
        self.dtype = self.meta.pop('dtype','f')
        self.ds_type = ds_type
        self._next_matrix = False
        self._save_timestamp = save_timestamp
        
        ## only one information: either 'name' (for creation) or 'ds_url' (for readout)
        if (name and ds_url) or (not name and not ds_url) :
            logging.error("HDF_dataset: Please specify [only] one, 'name' or 'ds_url' ")
            raise NameError
        if name:
            self._new_ds_defaults(name, unit, folder, comment)
        elif ds_url:
            self._read_ds_from_hdf(ds_url)

    def _new_ds_defaults(self, name, unit, folder, comment):
        self.name = name
        self.folder = folder
        self.ds_url = "/entry/" + folder + "0/" + name
        self.unit = unit
        self.comment = comment
        # the first dataset is used to extract a few attributes
        self.first = True

    def _read_ds_from_hdf(self,ds_url):
        ds = self.hf[str(ds_url)]

        for attr in ds.attrs.keys():
            val = ds.attrs.get(attr)
            setattr(self,attr,val)
        
        self.ds_url =  ds_url
        
    def _setup_metadata(self):
        ds = self.ds
        ds.attrs.create('ds_type',self.ds_type)            
        ds.attrs.create("comment",self.comment.encode())
        ds.attrs.create('ds_url',self.ds_url.encode())
        if self.ds_type != ds_types['txt']:
            ds.attrs.create('unit', self.unit.encode())
            if self.x_object:
                ds.attrs.create("x_ds_url",self.x_object.ds_url.encode())
            if self.y_object:
                ds.attrs.create("y_ds_url",self.y_object.ds_url.encode())
            if self.z_object:
                ds.attrs.create("z_ds_url",self.z_object.ds_url.encode())


    def next_matrix(self):
        self._next_matrix = True
        self._y_pos = 0
        
    def append(self,data, reset=False, pointwise=False):
        """Function to save a growing measurement dataset to the hdf file.
        
        Data is added one datapoint (vector) or one dataline (matrix, box) at a
        time. The data is cast to numpy arrays if possible and appended to the
        existing dataset. A timestamp-dataset is also recorded here.

        Args:
            data: any data to be appended to the dataset
            reset (Boolean, optional); indicator for appending, or resetting the dataset
            pointwise (Boolean): if True, the data is appended pointwise, i.e. to the innermost dimension
        """
        if self.ds_type == ds_types['txt']:
            try:
                data = data.encode("utf-8")
            except AttributeError:
                import json
                data = json.dumps(data, cls=QkitJSONEncoder, indent = 4, sort_keys=True)
        else:
            ## we cast everything to a float numpy array
            data = numpy.atleast_1d(numpy.array(data,dtype=self.dtype))
        # at this point the reference data should be around
        if self.first:
            self.first = False
            if self.ds_type == ds_types['txt']:
                tracelength = 0
            else:
                tracelength = len(data)
            ## tracelength is used so far only for multi-dimensional datasets to chunk needed memory
            self.ds = self.hf.create_dataset(self.name,tracelength,
                                             folder=self.folder,
                                             dim = self.dim,
                                             ds_type = self.ds_type,
                                             dtype = self.dtype,
                                             **self.meta)
            self._setup_metadata()
            if self._save_timestamp:
                self._create_timestamp_ds()

        self.hf.append(self.ds, data, next_matrix=self._next_matrix, reset=reset, pointwise=pointwise)
        if self._save_timestamp:
            self.hf.append(self.ds_ts, numpy.array([time.time()]), next_matrix=self._next_matrix, reset=reset)
        if self._next_matrix:
            self._next_matrix = False

        self.hf.flush()
            
    def add(self,data):
        """Function to save a 1dim dataset once.
        
        The add method is used to add data to a coordinate dataset. The data is
        cast to a numpy array.
        """

        if self.ds_type == ds_types['coordinate'] or self.ds_type == ds_types['vector']:
            self.append(data, reset = True)
        else:
            logging.info("add is only for 1-Dim data. Please use append 2-Dim data.")
            return False
        
    def _create_timestamp_ds(self):
        """Create datasets containing timestamps assocciated with a measurement dataset.
        
        A dataset with the same name (+ suffix '_ts') and dimension-1 is created
        and unix timestamp added at each append() call.
        """
        ds_type = {
            ds_types['vector']: ds_types['vector'],
            ds_types['matrix']: ds_types['vector'],
            ds_types['box']: ds_types['matrix']
        }[self.ds_type]
        self.ds_ts = self.hf.create_dataset(self.name + '_ts', tracelength=1, folder=self.folder, dim=max(self.dim - 1, 1), dtype='float64', ds_type=ds_type)
        self.ds_ts.attrs.create('name', 'measurement_time'.encode())       
        self.ds_ts.attrs.create('unit', 's'.encode())
        if self.ds_type == ds_types['matrix']:
            self.ds_ts.attrs.create("x_ds_url",self.ds.attrs.get('x_ds_url', ''))
        if self.ds_type == ds_types['box']:
            self.ds_ts.attrs.create("x_ds_url",self.ds.attrs.get('x_ds_url', ''))
            self.ds_ts.attrs.create("y_ds_url",self.ds.attrs.get('y_ds_url', ''))
    """
    def __getitem__(self, name):
        return self.hf[name]

    def __setitem__(self, name, val):
        self.hf[name] = val
    """
    def __repr__(self):
        ret = "HDF5Data '%s'" % (self.name)
        return ret



# HR@KIT 2017 / R.Heeres 2008
import qkit
import os
import sys
import logging

def _set_insdir():
    instdir = qkit.cfg.get('instruments_dir', None)
    if instdir and os.path.exists(instdir):

        sys.path.append(instdir)
        qkit.cfg['instruments_dir'] = instdir
        logging.info('Set instruments dir to %s'% instdir)
        return instdir
    else:
        qkit.cfg['instruments_dir'] = None
        logging.warning(__name__ + ' : "%s" is not a valid path for instruments_dir, setting to None' % instdir)  
        return None

def _set_user_insdir():
    '''
    Setting directory for user-specific instruments.
    For this config['user_instruments_dir'] needs to be defined.
    '''

    instdir = qkit.cfg.get('user_instruments_dir', None)

    if instdir and os.path.exists(instdir):
        absdir = os.path.abspath(instdir)
        qkit.cfg['user_instruments_dir'] = absdir
        logging.info('Set user instruments dir to %s'% absdir)
    else:
        qkit.cfg['user_instruments_dir'] = None
        logging.info(__name__ + ' : "%s" is not a valid path for user_instruments_dir, setting to None' % instdir)
        return None

    # old code below, may not be needed anymore
    if sys.path.count(absdir) != 0:
        return absdir
    else:
        idx = sys.path.index(_insdir)
        sys.path.insert(idx, absdir)
        return absdir


_insdir = _set_insdir()
_user_insdir = _set_user_insdir()
